import os
import pydicom
import argparse
import nibabel
import re
from typing import Union
from pathlib import Path
import contextlib
import sys
import json
from joblib import Parallel, delayed


try:
    import helper_functions
    import ecat
    import dcm2niix4pet
    import pet_metadata
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions
    import pypet2bids.ecat as ecat
    import pypet2bids.dcm2niix4pet as dcm2niix4pet
    import pypet2bids.pet_metadata as pet_metadata


def spread_sheet_check_for_pet(sourcefile: Union[str, Path], **kwargs):
    # load data from spreadsheet
    data = helper_functions.open_meta_data(sourcefile)

    try:
        pet_field_requirements = pet_metadata.PET_metadata
    except:
        pet_field_requirements = {}

    mandatory_fields = pet_field_requirements.get("mandatory", [])
    recommended_fields = pet_field_requirements.get("recommended", [])
    optional_fields = pet_field_requirements.get("optional", [])
    blood_recording_fields = pet_field_requirements.get("blood_recording_fields", [])

    intersection = set(
        mandatory_fields + recommended_fields + optional_fields + blood_recording_fields
    ) & set(data.keys())

    if len(intersection) > 0:
        return True
    else:
        return False


def read_files_in_parallel(file_paths: list, function, n_jobs=-2, **kwargs):
    """
    Read files in parallel using joblib (note this should be refactored to use the threading module)
    :param file_paths: list of file paths to read
    :param function: function to apply to each file
    :param n_jobs: number of jobs to run in parallel
    :param kwargs: keyword arguments to pass to function
    :return: list of results
    """
    # TODO replace dependency on joblib with threading module
    results = Parallel(n_jobs=n_jobs)(
        delayed(function)(file_path, **kwargs) for file_path in file_paths
    )
    return results


class DummyFile(object):
    def write(self, x):
        pass


@contextlib.contextmanager
def nostdout():
    save_stdout = sys.stdout
    sys.stdout = DummyFile()
    yield
    sys.stdout = save_stdout


def pet_file(file_path: Path, return_only_path=False) -> Union[bool, str]:
    """
    Given a file path determine if the file is a pet imaging or pet spreadsheet type of file.
    Returns a tuple with the 'PET' status of the file followed by the type of PET file, one of
    the following -> 'DICOM', 'ECAT', 'SPREADSHEET', ''

    True value and DICOM returned if a dicom file is found
    >>> status, type_of_pet_file = pet_file('PETDICOM001.img')
    >>> assert status == True
    >>> assert type_of_pet_file == 'DICOM'

    False value and empty string are returned if not a pet file e.g.:
    >>> status, type_of_pet_file = pet_file('is_pet.py')
    >>> assert status == False
    >>> assert type_of_pet_file == ''

    :param file_path: path to file to check
    :type file_path: pathlib.Path object
    :return: (status, file type)
    :rtype: tuple(bool, str)
    """
    if not file_path.exists():
        raise FileNotFoundError(file_path)

    file_type = ""
    # get suffix of file
    suffix = file_path.suffix

    # if there are suffixes, join them together
    if len(file_path.suffixes) > 1:
        suffix = "".join(file_path.suffixes)

    # suppress all stdout from other functions
    with nostdout():
        if not file_type and (
            suffix.lower() in [".dcm", ".ima", ".img", ""]
            or "mr" in str(file_path.name).lower()
            or bool(re.search(r"\d", suffix.lower()))
        ):
            try:
                read_file = pydicom.dcmread(file_path)
                if read_file.Modality == "PT":
                    file_type = "DICOM"
                else:
                    # do nothing, we only want dicoms with the correct modality
                    pass
            except (pydicom.errors.InvalidDicomError, AttributeError):
                pass

        if not file_type and suffix.lower() in [".v", ".v.gz"]:
            try:
                read_file = ecat.Ecat(str(file_path))
                file_type = "ECAT"
            except nibabel.filebasedimages.ImageFileError:
                pass

        if not file_type and suffix.lower() in [".xlsx", ".tsv", ".csv", ".xls"]:
            try:
                read_file = spread_sheet_check_for_pet(file_path)
                if read_file:
                    # if it looks like a pet file
                    file_type = "SPREADSHEET"
            except (IOError, ValueError):
                pass

    if file_type:
        if return_only_path:
            return Path(file_path)
        else:
            return True, file_type
    else:
        if return_only_path:
            pass
        else:
            return False, file_type


def pet_folder(folder_path: Path, skim=False, njobs=2) -> Union[str, list, bool]:
    if not folder_path.exists():
        raise FileNotFoundError(folder_path)
    if not folder_path.is_dir():
        raise FileNotFoundError(folder_path)

    all_files = []
    # collect list of all files
    for root, folders, files in os.walk(folder_path):
        for f in files:
            f_path = Path(os.path.join(root, f))
            if f_path.exists():
                all_files.append(Path(os.path.join(root, f)))

    # we aren't going to want to inspect every single file, instead we're going skim through the list of files
    # by first selecting only the first file of each folder that has a given suffix (especially for dicom files)
    if skim:
        from pprint import pprint

        # collect all the dicom images
        dicoms, spreadsheets, ecats = {}, {}, {}
        for f in all_files:
            if f.suffix.lower() in [".dcm", ".ima", ".img", ""]:
                parent = dicoms.get(str(f.parent), {str(f.parent): {f.suffix: f}})
                dicoms[str(f.parent)] = parent
            if f.suffix.lower() in [".xlsx", ".tsv", ".csv", ".xls"]:
                parent = spreadsheets.get(str(f.parent), {str(f.parent): {f.suffix: f}})
                spreadsheets[str(f.parent)] = parent
            if f.suffix.lower() in [".v"] or f.suffixes == [".v", ".gz"]:
                parent = ecats.get(
                    str(f.parent), {str(f.parent): {"".join(f.suffixes): f}}
                )
                ecats[str(f.parent)] = parent
        # now flatten all the dictionaries to only include the file parts
        smaller_list = []
        for file_dict in [dicoms, spreadsheets, ecats]:
            for k, v in file_dict.items():
                for k2, v2 in v.items():
                    for k3, v3 in v2.items():
                        smaller_list.append(v3)

        all_files = smaller_list

    # check if any files are pet files
    files = read_files_in_parallel(
        all_files, pet_file, n_jobs=njobs, return_only_path=True
    )
    files = [Path(f) for f in files if f is not None]
    # check through list of pet files and statuses for True values in parallel
    folders = set([f.parent for f in files])

    return folders


def main():
    """
    This command line utility exists almost entirely for ezBIDS. It's use there is to ensure that dcm2niix is not run
    on folders containing PET images. Instead, the PET images are converted using the dcm2niix4pet from this library.
    :return: Path to PET folder or file
    :rtype: str
    """

    parser = argparse.ArgumentParser(
        description="Check if a file is a PET image or bloodfile. If a folder is given, "
        "all files in the folder will be checked and any folders "
        "containing PET files will be returned."
    )
    parser.add_argument(
        "filepath",
        type=Path,
        help="File path to check whether file is PET image or bloodfile. "
        "If a folder is given, all files in the folder will be checked and "
        "any folders containing PET files will be returned.",
    )
    parser.add_argument(
        "-p",
        "--path-only",
        action="store_true",
        default=False,
        help="Omit type of pet file; only return file path if file is PET file",
    )
    parser.add_argument(
        "-s",
        "--skim",
        action="store_true",
        default=False,
        help="Only check files that "
        "are suspected to be PET files. Defaults to checking every file found in a folder."
        "When selected checks only a single file in folder ending in an extension that may be a PET FILE.",
    )
    parser.add_argument(
        "-n",
        "--njobs",
        type=int,
        default=2,
        help="Number of jobs to run in parallel when examining folders, defaults to 2.",
    )
    args = parser.parse_args()

    if args.filepath.is_file():
        status, pet_files = pet_file(args.filepath.resolve())
        if status:
            if args.path_only:
                print(f"{args.filepath}")
            else:
                print(f"{args.filepath} {pet_files}")
        else:
            sys.exit(1)

    elif args.filepath.is_dir():
        pet_folders = pet_folder(
            args.filepath.resolve(), skim=args.skim, njobs=args.njobs
        )
        if len(pet_folders) > 0:
            for f in pet_folders:
                print(f"{f}")
        else:
            sys.exit(1)
    else:
        sys.exit(1)


if __name__ == "__main__":
    main()



try:
    import ecat
    import write_ecat
    import read_ecat
    import helper_functions
except ImportError:
    from pypet2bids.ecat import ecat
    from pypet2bids.write_ecat import write_ecat
    from pypet2bids.read_ecat import read_ecat
    from pypet2bids import helper_functions

# collect ecat header jsons
ecat_headers = read_ecat.ecat_header_maps.get("ecat_headers")


def update_ecat_header(ecat_file: str, new_values: dict):
    """
    Update the header of an ECAT file with new values
    :param ecat_file: path to the ECAT file
    :param new_values: dictionary of new values to update
    :param ecat_header: dictionary of the ECAT header
    :return: None
    """

    # read ecat and determine version of ecat file
    print(f"Reading ECAT file {ecat_file}")
    infile = ecat.Ecat(ecat_file)

    # collect the appropriate header schema
    sw_version = str(infile.ecat_header.get("SW_VERSION", 73))

    infile_header_map = ecat_headers[sw_version]["mainheader"]

    # iterate through new values and update the header
    for name, value in new_values.items():
        if infile.ecat_header.get(name):
            if type(infile.ecat_header[name]) == type(value):
                infile.ecat_header[name] = value
            else:
                print(
                    f"WARNING: {name} has type {type(infile.ecat_header[name])} "
                    f"and new value {value} has type {type(value)}"
                )
        else:
            print(
                f"WARNING: {name} not found in header schema for ECAT {ecat_headers.ecat_header.get('SW_VERSION', 73)} "
                f"not updating with value {value}"
            )

    # update the header of the ecat file in question
    with open(infile.ecat_file, "r+b") as outfile:
        write_ecat.write_header(
            ecat_file=outfile, schema=infile_header_map, values=infile.ecat_header
        )


def cli():
    import argparse

    parser = argparse.ArgumentParser(description="Update the header of an ECAT file.")
    parser.add_argument("ecat_file", type=str, help="path to the ECAT file")
    parser.add_argument(
        "new_values",
        nargs="*",
        action=helper_functions.ParseKwargs,
        default={},
        help="new values to update the MAINHEADER of the ecat file, e.g. NUM_FRAMES=71 "
        "or CALIBRATION_FACTOR=0.5. "
        'or STUDY_DESCRIPTION="very important work"'
        "If the value is a string, it must be in quotes.",
    )
    args = parser.parse_args()
    update_ecat_header(ecat_file=args.ecat_file, new_values=args.new_values)


if __name__ == "__main__":
    cli()



"""
Simple command line tool to extract header and pixel information from ecat files and convert ecat to nifti.

| *Authors: Anthony Galassi*
| *Copyright OpenNeuroPET team*
"""

import argparse
import os
import pathlib
import sys
import textwrap
from os.path import join
from importlib.metadata import version

try:
    import helper_functions
    import Ecat
    from update_json_pet_file import check_json, check_meta_radio_inputs
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions
    from pypet2bids.ecat import Ecat
    from pypet2bids.update_json_pet_file import check_json, check_meta_radio_inputs

epilog = textwrap.dedent(
    """
    
    example usage:
    
    ecatpet2bids ecatfile.v --dump # dumps ecat header information
    ecatpet2bids ecatfile.v --json # dumps header and subheader information to stdout
    ecatpet2bids ecatfile.v --nifti sub-01_ses-example_pet.nii --convert --kwargs TimeZero="12:34:56" # convert to nii
    ecatpet2bids ecatfile.v --scannerparams ge_parameters.txt --nifti sub-01_pet.nii --convert # load scanner specific \
params
    
    For additional (highly verbose) example usage call this program with the --show-examples flag.
"""
)


def cli():
    """
    Builds an argparse.ArgumentParser() object to access the methods available in the Ecat class in pypet2bids.ecat.ECAT

    :param ecat: original ecat image to inspect or convert
    :type ecat: path
    :param --affine: display the affine matrix of the ecat
    :type --affie: stdout
    :param --convert: fattempt to convert the ecat file into a nifti defaults to False if flag isn't present
    :type --convert: flag
    :param --dump: dump the main header of the ecat file
    :type --dump: flag
    :param --json: output the entire header, subheaders, and affine matrix to stdout as jsan
    :type --json: flag
    :param --nifti: Name of the output nifti file
    :type --nifti: path
    :param --subheader: display just the subheaders to the stdout
    :type --subheader: flag
    :param --sidecar: output a bids formatted sidecar with the nifti, defaults to True
    :type --sidecar: flag
    :param --kwargs: include additional key/pair arguments to append to a sidecar file post conversion to nifti
    :type --kwargs: strings, nargs
    :param --scannerparams: a parameter.txt file to extract scanner specific kwargs/args/BIDS fields from; constant per scanner
    :type --scanerparamas: path
    :param --directory_table: collect directory table from ECAT, useful for poking around the ecat file bytewise by frame
    :type --director_table: flag
    :param --show-examples: shows verbose example usage of this cli
    :type --show-examples: flag
    :param --metadata-path: path to a spreadsheet containing PET metadata
    :type --metadata-path: path
    :param --ezbids: perform additional actions for ezbids
    :type --ezbids: flag

    :return: argparse.ArgumentParser.args for later use in executing conversions or ECAT methods
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=epilog,
        description="Extracts information from an ECAT file and/or converts ECAT imaging"
        " and PET metadata files to BIDS compliant nifti, json, and tsv",
    )
    update_or_convert = parser.add_mutually_exclusive_group()
    parser.add_argument(
        "ecat", nargs="?", metavar="ecat_file", help="Ecat image to collect info from."
    )
    parser.add_argument(
        "--affine", "-a", help="Show affine matrix", action="store_true", default=False
    )
    update_or_convert.add_argument(
        "--convert",
        "-c",
        required=False,
        action="store_true",
        help="If supplied will attempt conversion.",
    )
    parser.add_argument(
        "--dump",
        "-d",
        help="Dump information in Header",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--json",
        "-j",
        action="store_true",
        default=False,
        help="""
        Output header and subheader info as JSON to stdout, overrides all other options""",
    )
    parser.add_argument(
        "--nifti",
        "-n",
        metavar="file_name",
        help="Name of nifti output file",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--subheader",
        "-s",
        help="Display subheaders",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--sidecar",
        action="store_true",
        help="Output a bids formatted sidecar for pairing with" "a nifti.",
    )
    parser.add_argument(
        "--kwargs",
        "-k",
        nargs="*",
        action=helper_functions.ParseKwargs,
        default={},
        help="Include additional values in the nifti sidecar json or override values extracted from "
        'the supplied nifti. e.g. including `--kwargs TimeZero="12:12:12"` would override the '
        "calculated TimeZero. Any number of additional arguments can be supplied after --kwargs "
        "e.g. `--kwargs BidsVariable1=1 BidsVariable2=2` etc etc."
        "Note: the value portion of the argument (right side of the equal's sign) should always"
        'be surrounded by double quotes BidsVarQuoted="[0, 1 , 3]"',
    )
    parser.add_argument(
        "--scannerparams",
        nargs="*",
        help="Loads saved scanner params from a configuration file following "
        "--scanner-params/-s if this option is used without an argument "
        "this cli will look for any scanner parameters file in the "
        "directory with the name *parameters.txt from which this cli is "
        "called.",
    )
    parser.add_argument(
        "--directory_table",
        "-t",
        help="Collect table/array of ECAT frame byte location map",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--show-examples",
        "-E",
        "--HELP",
        "-H",
        help="Shows example usage of this cli.",
        action="store_true",
    )
    parser.add_argument(
        "--metadata-path", "-m", help="Path to a spreadsheet containing PET metadata."
    )
    update_or_convert.add_argument(
        "--update",
        "-u",
        type=str,
        default="",
        help="Update/create a json sidecar file from an ECAT given a path to that each "
        "file,. e.g."
        "ecatpet2bids ecatfile.v --update path/to/sidecar.json "
        "additionally one can pass metadata to the sidecar via inclusion of the "
        "--kwargs flag or"
        "the --metadata-path flag. If both are included the --kwargs flag will "
        "override any"
        "overlapping values in the --metadata-path flag or found in the ECAT file \n"
        "ecatpet2bids ecatfile.v --update path/to/sidecar.json --kwargs "
        'TimeZero="12:12:12"'
        "ecatpet2bids ecatfile.v --update path/to/sidecar.json --metadata-path "
        "path/to/metadata.xlsx",
    )
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=f"{helper_functions.get_version()}",
    )
    parser.add_argument(
        "--notrack",
        action="store_true",
        default=False,
        help="Opt-out of sending tracking information of this run to the PET2BIDS developers. "
        "This information helps to improve PET2BIDS and provides an indicator of real world "
        "usage crucial for obtaining funding.",
    )
    parser.add_argument(
        "--ezbids",
        default=False,
        action="store_true",
        help="Enable or disable extra steps performed for ezBIDS.",
    )

    return parser


example1 = textwrap.dedent(
    """
    
Usage examples are below, the first being the most brutish way of injecting BIDS required fields
into the output from ecatpet2bids. Additional arguments/fields are passed via the kwargs flag
in key value pairs.

example 1 (Passing PET metadata via the --kwargs argument):
    
    # Note `#` denotes a comment
    
    # ecatfile -> SiemensHRRT-NRU/XCal-Hrrt-2022.04.21.15.43.05_EM_3D.v  
    # nifti -> sub-SiemensHRRTNRU/pet/sub-SiemensHRRTNRU_pet.nii
    # kwargs -> a bunch of key pair arguments spaced 1 space apart with the values surrounded by double quotes

    ecatpet2bids SiemensHRRT-NRU/XCal-Hrrt-2022.04.21.15.43.05_EM_3D.v 
    --nifti sub-SiemensHRRTNRU/pet/sub-SiemensHRRTNRU_pet.nii --convert 
    --kwargs
    TimeZero="10:10:10"
    Manufacturer=Siemens
    ManufacturersModelName=HRRT
    InstitutionName="Rigshospitalet, NRU, DK"
    BodyPart=Phantom Units="Bq/mL"
    Manufacturer=Siemens
    ManufacturersModelName=HRRT
    InstitutionName="Rigshospitalet, NRU, DK"
    BodyPart="Phantom"
    Units="Bq/mL"
    TracerName=FDG
    TracerRadionuclide=F18
    InjectedRadioactivity=81.24
    SpecificRadioactivity="1.3019e+04"
    ModeOfAdministration="infusion"
    InjectedMass=1
    InjectedMassUnits=grams
    AcquisitionMode="list mode"
    ImageDecayCorrected="True"
    ImageDecayCorrectionTime=0
    ReconFilterType="None"
    ReconFilterSize=0
    AttenuationCorrection="10-min transmission scan"
    SpecificRadioactivityUnits="Bq"
    ScanStart=0
    InjectionStart=0
    InjectedRadioactivityUnits="Bq"
    ReconFilterType="['n/a']"
"""
)


def main():
    """
    Executes cli() and then uses Ecat class to convert or extract information from an Ecat file

    :return: N/A
    """
    cli_parser = cli()

    if len(sys.argv) == 1:
        cli_parser.print_usage()
        sys.exit(1)
    else:
        cli_args = cli_parser.parse_args()

    if cli_args.show_examples:
        print(example1)
        sys.exit(0)

    if cli_args.notrack:
        os.environ["PET2BIDS_TELEMETRY_ENABLED"] = "False"

    collect_pixel_data = False
    if cli_args.convert or cli_args.update or cli_args.sidecar:
        collect_pixel_data = True
    if cli_args.scannerparams is not None:
        # if no args are supplied to --scannerparams/-s
        if cli_args.scannerparams == []:
            files_in_command_line_dir_call = os.listdir()
            scanner_txt = None
            for each in files_in_command_line_dir_call:
                if "parameters.txt" in each:
                    scanner_txt = each
                    break
            if scanner_txt is None:
                called_dir = os.getcwd()
                error_string = (
                    f"No scanner file found in {called_dir}. Either create a parameters.txt file, omit "
                    f"the --scannerparams argument, or specify a full path to a scanner.txt file after the "
                    f"--scannerparams argument."
                )
                raise Exception(error_string)
        else:
            scanner_txt = cli_args.scannerparams[0]
        scanner_params = helper_functions.load_vars_from_config(scanner_txt)

        # if any additional non-null values have been included in a scanner.txt include those in the sidecar,
        # variable supplied via the --kwargs argument to the cli will override any variables in scanner.txt
        if scanner_params:
            scanner_params.update(cli_args.kwargs)
            cli_args.kwargs.update(scanner_params)

    ecat = Ecat(
        ecat_file=cli_args.ecat,
        nifti_file=cli_args.nifti,
        collect_pixel_data=collect_pixel_data,
        metadata_path=cli_args.metadata_path,
        kwargs=cli_args.kwargs,
        ezbids=cli_args.ezbids,
    )
    if cli_args.json:
        ecat.json_out()
        sys.exit(0)

    if cli_args.dump:
        ecat.show_header()
        sys.exit(0)
    if cli_args.directory_table:
        ecat.show_directory_table()
        sys.exit(0)
    if cli_args.affine:
        ecat.show_affine()
    if cli_args.subheader:
        ecat.show_subheaders()
    if cli_args.sidecar:
        ecat.populate_sidecar(**cli_args.kwargs)
        ecat.show_sidecar()
    if cli_args.convert:
        ecat.convert()
    if cli_args.update:
        ecat.update_pet_json(cli_args.update)


def update_json_with_ecat_value_cli():
    """
    Updates a json sidecar with values extracted from an ecat file, optionally additional values can be included
    via the -k --additional-arguments flag and/or a metadata spreadsheet can be supplied via the --metadata-path flag.
    Command can be accessed after installation via `upadatepetjsonfromecat`
    """
    json_update_cli = argparse.ArgumentParser(
        description="Updates a json sidecar with values extracted from an ECAT."
    )
    json_update_cli.add_argument(
        "-j", "--json", help="Path to a json to update file.", required=True
    )
    json_update_cli.add_argument(
        "-e", "--ecat", help="Path to an ecat file.", required=True
    )
    json_update_cli.add_argument(
        "-m", "--metadata-path", help="Path to a spreadsheet containing PET metadata."
    )
    json_update_cli.add_argument(
        "-k",
        "--additional-arguments",
        nargs="*",
        action=helper_functions.ParseKwargs,
        default={},
        help="Include additional values in the sidecar json or override values extracted "
        "from the supplied ECAT or metadata spreadsheet. "
        'e.g. including `--kwargs TimeZero="12:12:12"` would override the calculated '
        "TimeZero."
        "Any number of additional arguments can be supplied after --kwargs e.g. `--kwargs"
        "BidsVariable1=1 BidsVariable2=2` etc etc."
        "Note: the value portion of the argument (right side of the equal's sign) should "
        'always be surrounded by double quotes BidsVarQuoted="[0, 1 , 3]"',
    )

    args = json_update_cli.parse_args()

    update_ecat = Ecat(
        ecat_file=args.ecat,
        nifti_file=None,
        collect_pixel_data=True,
        metadata_path=args.metadata_path,
        kwargs=args.additional_arguments,
    )
    update_ecat.update_pet_json(args.json)

    # lastly check the json
    check_json(args.json, logger="check_json", silent=False)


if __name__ == "__main__":
    main()



from os.path import join
import warnings
import pandas
import pathlib

try:
    import helper_functions
    import pet_metadata
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions
    import pypet2bids.pet_metadata as pet_metadata


parent_dir = pathlib.Path(__file__).parent.resolve()
project_dir = parent_dir.parent

"""

:format:
:param:
:return:

Anthony Galassi
-----------------------------
Copyright Open NeuroPET team
"""


def translate_metadata(metadata_dataframe, image_path=NotImplemented):
    """
    This method exists as an example/template for an individual to customize and use on their own PET metadata contained
    within a spreadsheet. It is important to note that everything but the method name `translate_metadata` and the
    dictionary structure it returns can be modified to extract, transform, and load data from a user's spreadsheet.
    That is to say keep this method named `translate_metadata` and return a dictionary of the form:
    {
        'nifti_json: {},
        'blood_json: {},
        'blood_tsv: <pandas.DataFrame> or a <str>,
    }

    :param metadata_dataframe:
    :param image_path:
    :return:
    """
    nifti_json = {
        "Manufacturer": "",
        "ManufacturersModelName": "",
        "Units": "",
        "TracerName": "",
        "TracerRadionuclide": "",
        "InjectedRadioactivity": 0,
        "InjectedRadioactivityUnits": "",
        "InjectedMass": 0,
        "InjectedMassUnits": "",
        "SpecificRadioactivity": 0,
        "SpecificRadioactivityUnits": "",
        "ModeOfAdministration": "",
        "TimeZero": 0,
        "ScanStart": 0,
        "InjectionStart": 0,
        "FrameTimesStart": [],
        "FrameDuration": [],
        "AcquisitionMode": "",
        "ImageDecayCorrected": "",
        "ImageDecayCorrectionTime": 0,
        "ReconMethodName": "",
        "ReconMethodParameterLabels": [],
        "ReconMethodParameterUnits": [],
        "ReconMethodParameterValues": [],
        "ReconFilterType": "",
        "ReconFilterSize": 0,
        "AttenuationCorrection": "",
        "InstitutionName": "",
        "InstitutionalDepartmentName": "",
    }

    for key in nifti_json.keys():
        try:
            nifti_json[key] = helper_functions.flatten_series(metadata_dataframe[key])
        except KeyError:
            warnings.warn(f"{key} not found in metadata extracted from spreadsheet")

    blood_json = {
        "PlasmaAvail": True,
        "WholeBloodAvail": True,
        "MetaboliteAvail": False,
        "DispersionCorrected": False,
        "time": {
            "Description": "Time in relation to time zero defined in _pet.json",
            "Units": "s",
        },
        "plasma_radioactivity": {
            "Description": "Radioactivity in plasma samples, measured by eye balling it.",
            "Units": "kBq/mL",
        },
        "whole_blood_radioactivity": {
            "Description": "Radioactivity in whole blood samples, measured by divining rod.",
            "Units": "kBq/mL",
        },
    }

    blood_tsv = {
        "time": [],
        "plasma_radioactivity": [],
        "whole_blood_radioactivity": [],
    }

    for key in blood_tsv.keys():
        try:
            blood_tsv[key] = helper_functions.flatten_series(metadata_dataframe[key])
        except KeyError:
            warnings.warn(f"{key} not found in metadata extracted from spreadsheet")

    # now transform the key value pairs in blood_tsv into a pandas dataframe object.
    blood_tsv = pandas.DataFrame.from_dict(blood_tsv)

    return {"nifti_json": nifti_json, "blood_json": blood_json, "blood_tsv": blood_tsv}



"""
This is a lazy way to avoid opening a json, simply import this file to collect your BIDS sidecar templates instead. This
is not a function or a true module. This is just two python dictionaries with keys and empty value pairs.

**Parameters**

* sidecar_template_full: a dict containing every PET BIDS field
* sidecar_template_short: a dict containing only required PET BIDS fields

**Returns**     sidecar_template_full, sidecar_template_short

*Anthony Galassi*
*Copyright OpenNeuroPET team*

"""

sidecar_template_full = {
    "Manufacturer": "",
    "ManufacturersModelName": "",
    "Units": "",
    "InstitutionName": "",
    "InstitutionAddress": "",
    "InstitutionalDepartmentName": "",
    "BodyPart": "",
    "TracerName": "",
    "TracerRadionuclide": "",
    "InjectedRadioactivity": "",
    "InjectedRadioactivityUnits": "",
    "InjectedMass": "",
    "InjectedMassUnits": "",
    "SpecificRadioactivity": "",
    "SpecificRadioactivityUnits": "",
    "ModeOfAdministration": "",
    "TracerRadLex": "",
    "TracerSNOMED": "",
    "TracerMolecularWeight": "",
    "TracerMolecularUnits": "",
    "InjectedMassPerWeight": "",
    "InjectedMassPerWeightUnits": "",
    "SpecificRadioactivityMeasTime": "",
    "MolarActivity": "",
    "MolarActivityUnits": "",
    "MolarActivityMeasTime": "",
    "InfusionRadioActivity": "",
    "InfusionStart": "",
    "InfusionSpeed": "",
    "InfusionSpeedUnits": "",
    "InjectedVolume": "",
    "Purity": "",
    "PharamceuticalName": "",
    "PharmaceuticalDoseAmount": [],
    "PharmaceuticalDoseUnits": "",
    "PharmaceuticalDoseRegimen": "",
    "PharmaceuticalDoseTime": [],
    "Anaesthesia": "",
    "TimeZero": "",
    "ScanStart": "",
    "InjectionStart": "",
    "FrameTimesStart": [],
    "FrameDuration": [],
    "ScanDate": "",
    "InjectionEnd": "",
    "AcquisitionMode": "",
    "ImageDecayCorrected": "",
    "ImageDecayCorrectionTime": "",
    "ReconMethodName": "",
    "ReconMethodParameterLabels": [],
    "ReconMethodParameterUnits": [],
    "ReconMethodParameterValues": [],
    "ReconFilterType": [],
    "ReconFilterSize": [],
    "AttenuationCorrection": "",
    "ReconMethodImplementationVersion": "",
    "AttenuationCorrectionMethodReference": "",
    "ScaleFactor": [],
    "ScatterFraction": [],
    "DecayCorrectionFactor": [],
    "PromptRate": [],
    "RandomRate": [],
    "SinglesRate": [],
}

sidecar_template_short = {
    "Manufacturer": "",
    "ManufacturersModelName": "",
    "Units": "",
    "TracerName": "",
    "TracerRadionuclide": "",
    "InjectedRadioactivity": "",
    "InjectedRadioactivityUnits": "",
    "InjectedMass": "",
    "InjectedMassUnits": "",
    "SpecificRadioactivity": "",
    "SpecificRadioactivityUnits": "",
    "ModeOfAdministration": "",
    "TimeZero": "",
    "ScanStart": "",
    "InjectionStart": "",
    "FrameTimesStart": [],
    "FrameDuration": [],
    "AcquisitionMode": "",
    "ImageDecayCorrected": "",
    "ImageDecayCorrectionTime": "",
    "ReconMethodName": "",
    "ReconMethodParameterLabels": [],
    "ReconMethodParameterUnits": [],
    "ReconMethodParameterValues": [],
    "ReconFilterType": [],
    "ReconFilterSize": [],
    "AttenuationCorrection": "",
}



import requests
import os
import pathlib
import subprocess
import time
import sys
import select
from dotenv import load_dotenv
from typing import Union


try:
    from helper_functions import get_version, load_vars_from_config, modify_config_file
except ModuleNotFoundError:
    # collect pet2bids version
    from pypet2bids.helper_functions import (
        get_version,
        load_vars_from_config,
        modify_config_file,
    )

pet2bids_config = load_vars_from_config()
telemetry_default_url = pet2bids_config.get(
    "TELEMETRY_URL", "http://openneuropet.org/pet2bids/"
)
# check environment variables as well as the config file
telemetry_enabled_env = os.getenv("PET2BIDS_TELEMETRY_ENABLED", True)
telemetry_enabled = pet2bids_config.get("TELEMETRY_ENABLED", True)

# if telemetry is disabled in the config file or the environment variable disable its use.
if telemetry_enabled_env is False or telemetry_enabled is False:
    telemetry_enabled = False

telemetry_notify_user = pet2bids_config.get("NOTIFY_USER_OF_TELEMETRY", False)


def telemetry_enabled(config_path=None):
    """
    Check if telemetry is enabled, if it isn't disabled in the .pet2bidsconfig file
    it will be considered enabled. One must opt out of tracking usage manually.
    :param config_path: The path to the config file
    :type config_path: Union[str, pathlib.Path]
    :return: Whether telemetry is enabled or not
    :rtype: bool
    """
    # load dcm2niix file
    if not config_path:
        config_file = pathlib.Path.home() / ".pet2bidsconfig"
    else:
        config_file = pathlib.Path(config_path)

    if config_file.exists():
        load_dotenv(dotenv_path=config_file)

    # check to see if telemetry is disabled
    if (
        os.getenv("PET2BIDS_TELEMETRY_ENABLED", "").lower() == "false"
        or os.getenv("CI", "false") == "true"
    ):
        return False
    else:
        return True


def send_telemetry(json_data: dict, url: str = telemetry_default_url):
    """
    Send telemetry data to the telemetry server, by default this will first try
    to load the telemetry server url from the config file, if it's not found it will
    default to the hardcoded value in this module. This will always send, unless a user
    has disabled the pet2bids telemetry in the .pet2bidsconfig file.
    :param json_data: The dat to be sent to the telemetry server
    :type json_data: dict
    :param url: The url of the telemetry server
    :type url: str
    """
    if telemetry_enabled():
        # update data with version of pet2bids
        json_data["pypet2bids_version"] = get_version()
        # check if it's one of the dev's running this
        running_from_cloned_repository = subprocess.run(
            ["git", "status"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        if running_from_cloned_repository.returncode == 0:
            json_data["running_from_cloned_repository"] = True

        json_data["description"] = "pet2bids_python_telemetry"

        try:
            # Send a POST request to the telemetry server
            requests.post(url, json=json_data)
        except requests.exceptions.RequestException as e:
            pass
    else:
        pass


def count_input_files(input_file_path: Union[str, pathlib.Path]):
    """
    Count the number of input files in the input file path
    :param input_file_path: location of the input files
    :type input_file_path: Union[str, pathlib.Path]
    :return: The number of input files and their size
    :rtype: dict
    """
    if isinstance(input_file_path, str):
        input_file_path = pathlib.Path(input_file_path)

    if input_file_path.is_dir():
        # collect all files in the input dir
        all_files = [f for f in input_file_path.iterdir()]
        # count the number of dicom files
        num_dicom_files = 0
        dicom_files_size = 0
        total_files = 0
        total_files_size = 0
        for file in all_files:
            if file.is_file():
                total_files += 1
                total_files_size += file.stat().st_size
    return {
        "TotalInputFiles": total_files,
        "TotalInputFilesSize": total_files_size,
    }


def count_output_files(output_file_path: Union[str, pathlib.Path]):
    """
    Count the number of output files in the output file path. This can be useful
    in determining whether any additional files were created during the conversion process.
    However, this is only useful if the conversion process initially takes place from within
    a temporary directory.

    # TODO check the last modified date of the files to determine if they were created during
    # TODO the conversion process. That should make this useful in all cases.
    :param output_file_path: location of the output files
    :type output_file_path: Union[str, pathlib.Path]
    :return: The number of output files and their size and specifically the number of nifti files and
    their size
    :rtype: dict
    """
    if isinstance(output_file_path, str):
        output_file_path = pathlib.Path(output_file_path)

    if output_file_path.is_file():
        output_file_path = output_file_path.parent

    if output_file_path.is_dir():
        # collect all files in the output dir
        all_files = [f for f in output_file_path.iterdir()]
        # count the number nifti files
        num_nifti_files = 0
        nifti_files_size = 0
        total_files = 0
        total_files_size = 0
        for f in all_files:
            if f.is_file():
                total_files += 1
                total_files_size += f.stat().st_size
                if str(f).endswith(".nii") or str(f).endswith(".nii.gz"):
                    num_nifti_files += 1
                    nifti_files_size += f.stat().st_size
    return {
        "TotalOutputFiles": total_files,
        "TotalOutputFilesSize": total_files_size,
        "NiftiFiles": num_nifti_files,
        "NiftiFilesSize": nifti_files_size,
    }



"""
This module converts PMOD blood files (extension .bld) into BIDS compliant tsv and json files. It contains a class
PmodToBlood and a cli that uses that class to interface with the user. This module currently exists solely to convert
PMOD blood files and is not incorporated within any of the other conversion tools in PET2BIDS.

The command line arguments are details below, for more information about how this module works see the documentation for
PmodToBlood.

| *Authors: Anthony Galassi*
| *Copyright: OpenNeuroPET team*
"""

import json
import logging
import textwrap

import numpy
import pandas
import pandas as pd
import argparse
import warnings
import re
import ast
from pathlib import Path
from os.path import join

# from pypet2bids.helper_functions import ParseKwargs, collect_bids_part, open_meta_data

try:
    import helper_functions
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions

epilog = textwrap.dedent(
    """
    
    example usage:
    
    convert-pmod-to-blood --whole-blood-path wholeblood.bld --parent-fraction parentfraction.bld # simplest use case
    convert-pmod-to-blood --whole-blood-path wholeblood.bld --parent-fraction parentfraction.bld --plasma-activity-path plasma.bld
    convert-pmod-to-blood --whole blood-path wholeblood.bld --parent-fraction parentfraction.bld --outputh-path sub-01/pet
    
    For more extensive examples rerun this program with the --show-example flag
"""
)

example1 = textwrap.dedent(
    """
Usage examples are below that verbosely describe and show the input consumed an the output generated by this program.
Additional arguments/fields are passed via the kwargs flag in key value pairs.

Note: lines prepended with # denote comments/notes where as lines without # denote data or input arguments

example 1 (passing the bare minimum):
    
    # running the following
    convert-pmod-to-blood -whole whole_blood.bld -parent plasma_parent.bld
    # will result in outputting a tsv like the following:
    
    time	whole_blood_radioactivity	metabolite_parent_fraction
    25.2	0.000885	0.000874
    43.2	0.0192	0.00603
    51	0.92	1.38
    63	18.34858412	26.44553306
    90	51.96910035	73.40036984
    109.2	59.02807247	83.01412114
    130.8	74.9568219	106.0071338
    148.2	88.02307965	124.3830225
    171	86.93918061	120.924026
    189	47.18244008	65.42298506
    214.8	28.06571066	38.95386252
    319.2	10.22288119	14.11052535
    613.8	6.586049482	7.581724865
    919.2	7.448797008	5.784621668
    1810.2	6.49534108	4.294965789
    2719.2	5.373595253	2.836715768
    3607.2	4.798934663	2.48197381
    5419.8	3.898890497	2.05348682
    7207.2	3.772252717	1.77176473
    
    # and a json data dictionary as well
    {
      "WholeBloodAvail": "true",
      "MetaboliteAvail": "true",
      "time": {
        "Description": "Time in relation to time zero defined by the _pet.json",
        "Units": "s"
      },
      "whole_blood_radioactivity": {
        "Description": "Radioactivity in whole blood samples. Measured using COBRA counter.",
        "Units": "kBq/mL"
      },
      "metabolite_parent_fraction": {
        "Description": "Parent fraction of the radiotracer",
        "Units": "arbitrary"
      }
    }

 
"""
)


def cli():
    """
    Command line interface used to collect arguments for PmodToBlood class, note parameters defined below apply to the
    invocation at the command line **not** the method cli itself which takes no argument in Python.

    :param whole-blood-path: path to a PMOD .bld file containing the whole blood activity of a subject/run/scan
    :param parent-fraction-path: path to a PMOD .bld file containing the parent fraction of a subject/run/scan
    :param plasma-activity-path: path to a PMOD .bld file containing the plasma activity of a subject/run/scan
    :param output-path: the desired output path for the converted tsv and json files, if the path contains BIDS subject id's
        and session id's these will be extracted from the path and inserted into the filenames of the resultant tsv and json
        files.
    :param json: create a json sidecar/data-dictionary file along with output tsv's, default is set to True
    :param engine: engine used to read excel files, ignore this option as it will most likely be deprecated in the future
    :param kwargs: additional key pair arguments one wishes to include, such as extra entries for plasma or blood PET BIDS
        fields that aren't in PMOD blood files
    :param show-examples: shows an example of how to run this module as well as the outputs
    :return: collected arguments
    :rtype: argparse.namespace

    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Converts PMOD blood files to PET BIDS compliant tsv's and json "
        "_blood.* files",
    )
    parser.add_argument(
        "--whole-blood-path",
        "-w",
        help="Path to pmod whole blood file.",
        required=False,
        type=Path,
    )
    parser.add_argument(
        "--parent-fraction-path",
        "-f",
        help="Path to pmod parent fraction path.",
        required=False,
        type=Path,
    )
    parser.add_argument(
        "--plasma-activity-path",
        "-p",
        help="Path to pmod plasma file.",
        required=False,
        type=Path,
        default=None,
    )
    parser.add_argument(
        "--output-path",
        "-o",
        help="""Output path for output files (tsv and json) provide an existing folder path, if the output path is a
         BIDS path containing subject id and session id those values will be extracted an used to name the output 
         files.""",
        type=Path,
        default=None,
    )
    parser.add_argument(
        "--json",
        "-j",
        help="Output a json data dictionary along with tsv files (default True)",
        default=True,
        type=bool,
    )
    parser.add_argument(
        "--engine",
        "-e",
        help="Engine for loading PMOD files, see options for pandas.read_excel. Defaults to None.",
        default="",
        type=str,
    )
    parser.add_argument(
        "--kwargs",
        "-k",
        nargs="*",
        action=helper_functions.ParseKwargs,
        help="Pass additional arguments not enumerated in this help menu, see documentation online"
        + " for more details.",
        default={},
    )
    parser.add_argument(
        "--show-examples",
        help="Show additional examples (verbose) of how to use this interface.",
        action="store_true",
    )

    args = parser.parse_args()

    return args


def type_cast_cli_input(kwarg_arg):
    """
    This method sanitizes collects inputs from the cli and casts them as native python data types.

    :param kwarg_arg: a kwarg argument parsed from the command line, literal string
    :type kwarg_arg: str
    :return: kwarg argument evaluated to python datatype
    :rtype: dict, list, str, int, float, bool
    """
    try:
        var = ast.literal_eval(kwarg_arg)
        if type(var) in [dict, list, str, int, float, bool]:
            return var
    except (ValueError, SyntaxError):
        # try truthy evals if the input doesn't evaluate as a python type listed in the try statement
        if kwarg_arg.lower() in ["true", "t", "yes"]:
            return True
        elif kwarg_arg.lower() in ["false", "f", "no"]:
            return False
        else:
            return kwarg_arg


class PmodToBlood:
    """
    Converts PMOD blood files to PET BIDS compliant tsv's and json _blood.* files

    :param whole_blood_activity: path to a PMOD whole blood activity file
    :type whole_blood_activity: pathlib.Path
    :param parent_fraction: path to a PMOD metabolite parent blood file
    :type parent_fraction: pathlib.Path
    :param output_path: path to write output tsv and jsons to, defaults to parent folder of whole blood input file
    :type output_path: pathlib.Path
    :param output_json: boolean specifying whether to output a json sidecar for the blood files, default is false
    :type output_json: bool
    :param engine: soon to be deprecated, determines what engine is used in pandas.read_excel
    :type engine: str
    :param kwargs: additional plasma/blood/radioactivity key/pair BIDS arguments, not required but if provided used during
        logic phase and written to sidecar json
    :type kwargs: dict
    """

    def __init__(
        self,
        whole_blood_activity: Path,
        parent_fraction: Path = Path(),
        plasma_activity: Path = Path(),
        output_path: Path = None,
        output_json: bool = False,
        engine="",
        **kwargs,
    ):

        if kwargs:
            try:
                self.kwargs = kwargs["kwargs"]
            except KeyError:
                self.kwargs = kwargs
        else:
            self.kwargs = {}

        # relative closeness when comparing dataframes/inputs, default is set below
        self.rtol = 0.01
        self.atol = 0.02

        # cast input from kwargs
        for key, value in self.kwargs.items():
            self.kwargs[key] = type_cast_cli_input(value)

        self.units = None
        self.engine = engine

        # if given an output name run with that, otherwise we construct a name from the parent path the .bld files were
        # found at.
        if output_path:
            self.output_path = Path(output_path)
            if not self.output_path.is_dir():
                raise FileNotFoundError(
                    f"The output_path {output_path} must be an existing directory."
                )
        else:
            self.output_path = Path(whole_blood_activity).parent

        # check the output name for subject and session id
        if helper_functions.collect_bids_part("sub", str(self.output_path)):
            self.subject_id = helper_functions.collect_bids_part(
                "sub", self.output_path
            )
        else:
            print("Subject id not found in output_path, checking key pair input.")
            self.subject_id = self.kwargs.get("subject_id", "")

        if helper_functions.collect_bids_part("ses", str(self.output_path)):
            self.session_id = helper_functions.collect_bids_part(
                "ses", self.output_path
            )
        else:
            print("Session id not found in output_path, checking key pair input.")
            self.session_id = self.kwargs.get("session_id", "")

        self.output_json = output_json

        self.auto_sampled = []
        self.manually_sampled = []
        self.blood_series = {}
        self.duplicates = (
            {}
        )  # list of times that are duplicated across manual and automatic samples

        if whole_blood_activity.is_file():
            self.blood_series["whole_blood_activity"] = self.load_pmod_file(
                whole_blood_activity, engine=self.engine
            )

        if parent_fraction.is_file():
            self.blood_series["parent_fraction"] = self.load_pmod_file(
                parent_fraction, engine=self.engine
            )

        self.multiply_plasma = False
        # plasma activity is not required, but is used if provided
        if plasma_activity.is_file():
            self.blood_series["plasma_activity"] = self.load_pmod_file(
                plasma_activity, engine=self.engine
            )
            if "whole" in str.lower(plasma_activity.name) and "ratio" in str.lower(
                plasma_activity.name
            ):
                self.multiply_plasma = True

        # one may encounter data collected manually and/or automatically, we vary our logic depending on the case
        self.data_collection = {}

        for blood_sample in self.blood_series.keys():
            var = f"{blood_sample}_collection_method"
            if not kwargs.get(var, None):
                self.ask_recording_type(blood_sample)
            else:
                self.data_collection[blood_sample] = kwargs.get(var)

        for measure, collection_method in self.data_collection.items():
            if collection_method == "manual":
                self.manually_sampled.append({"name": measure})
            if collection_method == "automatic":
                self.auto_sampled.append({"name": measure})

        # scale time to seconds rename columns
        self.scale_time_rename_columns()

        # check blood files for consistency
        self.check_time_info()

        # warn if plasma data may be interpolated
        self.check_for_interpolated_data()

        # remove duplicates from automatically sampled data, this data will then be written out to manual versions
        # of each recording
        # self.deduplicate()

        self.write_out_tsvs()

        if self.output_json:
            self.write_out_jsons()

    @staticmethod
    def load_pmod_file(pmod_blood_file: Path, engine=""):
        """
        Loads a pmod .bld blood file in with pandas.

        :param pmod_blood_file: path to pmod blood file
        :type pmod_blood_file: pathlib.Path
        :param engine: python engine used to read excel sheet with pandas.read_excel,
        :type engine: str
        :return: contents of .bld file
        :rtype: pandas.DataFrame
        """
        if pmod_blood_file.is_file() and pmod_blood_file.exists():
            loaded_file = helper_functions.open_meta_data(pmod_blood_file)
            # drop unnamed columns
            loaded_file = loaded_file.loc[
                :, ~loaded_file.columns.str.contains("^Unnamed")
            ]
            return loaded_file
        else:
            raise FileNotFoundError(str(pmod_blood_file))

    def check_time_info(self):
        """
        Checks for time units, and time information between .bld files, number of rows and the values
        in the time index must be the same across each input .bld file. Additionally, renames time column
        to 'time' instead of what it's defined as in the pmod file. Does some more checks and scaling of plasma data,
        de-duplication etc.
        """
        # if there is only a single input do nothing, else go through each file. This shouldn't get reached
        # as whole_blood_activity and plasma_activity are required
        if len(self.blood_series) >= 2 and len(set(self.data_collection.values())) == 1:
            row_lengths = {}
            for key, bld_data in self.blood_series.items():
                row_lengths[key] = len(bld_data)

            if len(set(row_lengths.values())) > 1:
                err_message = (
                    f"Sampling method for all PMOD blood files (.bld) given as "
                    f"{list(set(self.data_collection.values()))[0]} must be of the same dimensions"
                    f" row-wise!\n"
                )
                for key, value in row_lengths.items():
                    err_message += f"{key} file has {value} rows\n"

                err_message += "Check input files are valid."

                raise Exception(err_message)

            # lastly make sure the same time points exist across each input file/dataframe
            whole_blood_activity = self.blood_series.pop("whole_blood_activity")
            for key, dataframe in self.blood_series.items():
                try:
                    assert whole_blood_activity["time"].equals(dataframe["time"])
                except AssertionError:
                    raise AssertionError(
                        f"Time(s) must have same values between input files, check time columns."
                    )
            # if it all checks out put the whole blood activity back into our blood series object
            self.blood_series["whole_blood_activity"] = whole_blood_activity

        # checks to make sure that an auto-sampled file has more entries in it than a manually sampled file,
        # John Henry must lose.
        elif (
            len(self.blood_series) >= 2 and len(set(self.data_collection.values())) > 1
        ):
            # check to make sure auto sampled .bld files have more entries than none autosampled
            compare_lengths = []
            for key, sampling_type in self.data_collection.items():
                compare_lengths.append(
                    {
                        "name": key,
                        "sampling_type": sampling_type,
                        "sample_length": len(self.blood_series[key]),
                    }
                )

            auto_sampled, manually_sampled = [], []
            for each in compare_lengths:
                if "auto" in str.lower(each["sampling_type"]):
                    auto_sampled.append(each)
                elif "manual" in str.lower(each["sampling_type"]):
                    manually_sampled.append(each)

            for auto in auto_sampled:
                for manual in manually_sampled:
                    if auto["sample_length"] < manual["sample_length"]:
                        logging.warning(
                            f"Autosampled .bld input for {auto['name']} has {auto['sample_length']} rows"
                            f" and Manually sampled input has {manual['sample_length']} rows. Autosampled blood "
                            f"files should have more rows than manually sampled input files. Check .bld inputs."
                        )

            # we want to make sure that in the case of mixed data we don't include manual time points in the autosampled
            # curves, first we find duplicates.

            whole_blood_activity = self.blood_series.get(
                "whole_blood_activity", pandas.DataFrame()
            )
            plasma_activity = self.blood_series.get(
                "plasma_activity", pandas.DataFrame()
            )
            parent_fraction = self.blood_series.get(
                "parent_fraction", pandas.DataFrame()
            )
            duplicates = {}

            # wp is short for whole blood and plasma activity, we're just using wp to make our conditional statements
            # look a bit cleaner while retaining the namespace from which activity series come from. Hanging onto these
            # names in this dictionary makes creating manual files later during de-duplication
            wp = {
                "whole_blood_activity": whole_blood_activity,
                "plasma_activity": plasma_activity,
            }

            # case 1 whole blood time == plasma time && parent fraction time exists
            if (
                len(wp["whole_blood_activity"]) == len(wp["plasma_activity"])
                and parent_fraction
            ):
                # get ready for duplicates
                for key in wp.keys():
                    if self.data_collection.get(key) == "automatic":
                        duplicates[key] = []

                for activity_series in duplicates.keys():
                    for pf_time in parent_fraction.get("time", []):
                        for activity_time in wp.get(key).get("time", []):
                            if numpy.isclose(pf_time, activity_time, atol=self.atol):
                                duplicates[activity_series].append(activity_time)

            # case 2 whole blood time > plasma time && parent fraction time exists
            elif (
                len(wp["whole_blood_activity"]) > len(wp["plasma_activity"])
                and parent_fraction is not None
            ):
                # get ready for duplicates
                for key in wp.keys():
                    if self.data_collection.get(key) == "automatic":
                        duplicates[key] = []

                if len(wp["plasma_activity"]) == (len(parent_fraction) - 1):
                    if (
                        parent_fraction["time"][0]
                        == wp["whole_blood_activity"]["time"][0]
                        and parent_fraction["time"][0] == 0
                    ):
                        if self.multiply_plasma:
                            wp["plasma_activity"].loc[-1] = [0, 1]
                        else:
                            wp["plasma_activity"].loc[-1] = [0, 0]

                        wp["plasma_activity"].index = wp["plasma_activity"].index + 1
                        wp["plasma_activity"] = (
                            wp["plasma_activity"]
                            .sort_values("time")
                            .reset_index(drop=True)
                        )
                        self.blood_series["plasma_activity"] = wp["plasma_activity"]
                    else:
                        raise Exception(
                            f"Parent fraction and plasma times do not match, we cannot figure out which "
                            f"time points to extract in whole blood."
                        )

                for activity_series in duplicates.keys():
                    for pf_time in parent_fraction.get("time", []):
                        for df_time in wp.get(key).get("time", []):
                            if numpy.isclose(pf_time, df_time, atol=self.atol):
                                duplicates[activity_series].append(df_time)

            # end else w/ no duplicates
            else:
                pass

            # catch only unique duplicates
            for k, v in duplicates.items():
                if v != []:
                    self.duplicates[k] = v

            # now we go ahead and remove any duplicates that exist in existing blood data
            for activity_series, duplicate_times in self.duplicates.items():
                # for name, sample_type in self.data_collection.items():
                #     if activity_series == name and sample_type == 'automatic':
                #         series_is_auto_sampled = True
                #     else:
                #         series_is_auto_sampled = False

                # we only want to remove rows if they exist in autosampled data, but are clearly manually sampled
                if self.data_collection.get(activity_series) == "automatic":
                    removed_rows = pandas.DataFrame(
                        columns=self.blood_series.get(
                            activity_series, pandas.DataFrame()
                        ).columns
                    )

                    for time in duplicate_times:
                        index = helper_functions.get_coordinates_containing(
                            time, self.blood_series[activity_series], exact=False
                        )

                        if index:
                            row = index[0][0]

                            dropped_row = helper_functions.drop_row(
                                self.blood_series[activity_series], row
                            )

                            # using the loc indexer, append the series to the end of the df
                            removed_rows.loc[len(removed_rows)] = dropped_row

                # update the blood series object with the popped/dropped row
                new_string = f"{activity_series}_manually_popped"
                self.blood_series[new_string] = removed_rows.sort_values(
                    "time"
                ).reset_index(drop=True)
                self.manually_sampled.append({"name": new_string})
                self.data_collection[new_string] = "manual"

                if len(self.blood_series[new_string]) == (len(parent_fraction) - 1):
                    if (
                        self.blood_series["parent_fraction"]["time"][0]
                        != self.blood_series[new_string]["time"][0]
                        and parent_fraction["time"][0] == 0
                    ):

                        self.blood_series[new_string].loc[-1] = [0, 0]
                        self.blood_series[new_string].index = (
                            self.blood_series[new_string].index + 1
                        )
                        self.blood_series[new_string] = (
                            self.blood_series[new_string]
                            .sort_values("time")
                            .reset_index(drop=True)
                        )

                    else:
                        raise Exception(
                            f"Parent fraction and plasma times do not match, we cannot figure out which "
                            f"time points to extract in whole blood."
                        )

        # if provided with the plasma ratio we multiply the whole blood by it, math time
        if self.multiply_plasma:
            # get our plasma activity, we get the manually popped if it exists
            pa = self.blood_series.get(
                "plasma_activity_manually_popped", self.blood_series["plasma_activity"]
            )

            if (
                self.blood_series.get(
                    "whole_blood_activity_manually_popped", pandas.DataFrame()
                ).shape[0]
                != 0
            ):

                wba = self.blood_series["whole_blood_activity_manually_popped"]
                new_plasma = (
                    wba["whole_blood_radioactivity"]
                    * self.blood_series["plasma_activity"]["plasma_radioactivity"]
                )

                temp = pandas.DataFrame(
                    {
                        "time": self.blood_series["plasma_activity"]["time"],
                        "plasma_radioactivity": new_plasma,
                    }
                )

                self.blood_series["plasma_activity"] = temp

            else:
                self.blood_series["plasma_activity"] = (
                    self.blood_series["plasma_activity"]
                    * self.blood_series["whole_blood"]
                )

    def scale_time_rename_columns(self):
        """
        Scales time info if it's not in seconds and renames dataframe column to 'time' instead of given column name in
        .bld file. Renames radioactivity column to BIDS compliant column name if it's in units Bq/cc or  Bq/mL.
        """
        # scale time info to seconds if it's minutes
        for name, dataframe in self.blood_series.items():
            time_column_header_name = []
            radioactivity_column_header_name = []
            parent_fraction_column_header_name = []
            time_scalar = 1.0
            time_column_header_name = [
                header
                for header in list(dataframe.columns)
                if "sec" in str.lower(header)
            ]
            if not time_column_header_name:
                time_column_header_name = [
                    header
                    for header in list(dataframe.columns)
                    if "min" in str.lower(header)
                ]
                if time_column_header_name:
                    time_scalar = 60.0

            if time_column_header_name and len(time_column_header_name) == 1:
                dataframe.rename(
                    columns={time_column_header_name[0]: "time"}, inplace=True
                )
            else:
                raise Exception(
                    "Unable to locate time column in blood file, make sure input files are formatted "
                    "to include a single time column in minutes or seconds."
                )

            # scale the time column to seconds
            dataframe["time"] = dataframe["time"] * time_scalar
            self.blood_series[name] = dataframe

            # locate parent fraction column
            parent_fraction_column_header_name = [
                header for header in dataframe.columns if "parent" in str.lower(header)
            ]
            parent_fraction_is_suspicous = False
            if len(parent_fraction_column_header_name) >= 1:
                pf = str(parent_fraction_column_header_name[0]).lower()
                if "bq" in pf or "ml" in pf:
                    logging.warning(
                        f"Found {parent_fraction_column_header_name[0]} in {name} input file, this column "
                        f"must be unitless"
                    )
                    parent_fraction_is_suspicous = True

            if not parent_fraction_column_header_name or parent_fraction_is_suspicous:
                # locate radioactivity column
                radioactivity_column_header_name = [
                    header
                    for header in dataframe.columns
                    if "bq" and "cc" in str.lower(header)
                ]
                # run through radio updating conversion if not percent parent
            if radioactivity_column_header_name and len(time_column_header_name) == 1:
                sub_ml_for_cc = re.sub("cc", "mL", radioactivity_column_header_name[0])
                extracted_units = re.search(r"\[(.*?)\]", sub_ml_for_cc)
                second_column_name = None
                if "plasma" in str.lower(radioactivity_column_header_name[0]) or (
                    "bq/cc" in str.lower(radioactivity_column_header_name[0])
                    and name == "plasma_activity"
                ):
                    second_column_name = "plasma_radioactivity"

                if name == "whole_blood_activity":
                    if "whole" in str.lower(
                        radioactivity_column_header_name[0]
                    ) or "blood" in str.lower(radioactivity_column_header_name[0]):
                        second_column_name = "whole_blood_radioactivity"

                if second_column_name:
                    dataframe = dataframe.rename(
                        {radioactivity_column_header_name[0]: second_column_name},
                        axis="columns",
                    )

                if extracted_units:
                    self.units = extracted_units.group(1)
                else:
                    raise Exception(
                        "Unable to determine radioactivity entries from .bld column name. Column name/units must be in "
                        "Bq/cc or Bq/mL"
                    )
            # if percent parent rename column accordingly
            if (
                parent_fraction_column_header_name
                and len(parent_fraction_column_header_name) == 1
                and name != "plasma_activity"
            ):
                dataframe = dataframe.rename(
                    {
                        parent_fraction_column_header_name[
                            0
                        ]: "metabolite_parent_fraction"
                    },
                    axis="columns",
                )
            self.blood_series[name] = dataframe.copy(deep=True)

    def ask_recording_type(self, recording: str):
        """
        Prompt user about data collection to determine how data was collected for each
        measure. e.g. auto-sampled, manually drawn, or a combination of the two.

        :param recording: the name of the recording
        :type: str
        :return: None
        :rtype: None
        """
        how = None
        while how != "a" or how != "m":
            how = input(
                f"How was the {recording} data sampled?:\nEnter A for automatically or M for manually\n"
            )
            if str.lower(how) == "m":
                self.data_collection[recording] = "manual"
                break
            elif str.lower(how) == "a":
                self.data_collection[recording] = "automatic"
                break
            elif str.lower(how) == "y":
                self.data_collection[recording] = "manual"
                warnings.warn(
                    f"Received {how} as input, assuming input received from cli w/ '-y' option on bash/zsh etc, "
                    f"defaulting to manual input"
                )
                break
            else:
                print(
                    f"You entered {how}; please enter either M or A to exit this prompt"
                )

    def write_out_tsvs(self):
        """
        Writes out blood data to tsv files corresponding to autosampled or manually sampled versions (dependent on user
        input)

        :return: None
        :rtype: None
        """
        # first we combine the various blood data into one or two dataframes, the autosampled data goes into a
        # recording_autosample, and the manually sampled data goes into a recording_manual if they exist
        if self.subject_id:
            file_path = join(self.output_path, self.subject_id + "_")
            if self.session_id:
                file_path += self.session_id + "_"
            manual_path = file_path + "recording-manual_blood.tsv"
            automatic_path = file_path + "recording-automatic_blood.tsv"
        else:
            manual_path = join(self.output_path, "recording-manual_blood.tsv")
            automatic_path = join(self.output_path, "recording-automatic_blood.tsv")

        # first combine autosampled data
        if self.auto_sampled:
            first_auto_sampled = self.blood_series[self.auto_sampled.pop()["name"]]
            for remaining_auto in self.auto_sampled:
                remaining_auto = self.blood_series[remaining_auto["name"]]
                column_difference = remaining_auto.columns.difference(
                    first_auto_sampled.columns
                )
                for column in list(column_difference):
                    first_auto_sampled[column] = remaining_auto[column]
            first_auto_sampled.to_csv(automatic_path, sep="\t", index=False)

        # combine any additional manually sampled dataframes
        if self.manually_sampled:
            first_manually_sampled = self.blood_series[
                self.manually_sampled.pop()["name"]
            ]
            for remaining_manual in self.manually_sampled:
                remaining_manual = self.blood_series[remaining_manual["name"]]
                column_difference = remaining_manual.columns.difference(
                    first_manually_sampled.columns
                )
                for column in list(column_difference):
                    first_manually_sampled[column] = remaining_manual[column]
            first_manually_sampled.to_csv(manual_path, sep="\t", index=False)

    def write_out_jsons(self):
        """
        Writes out sidecar json to correspond to _blood.tsv files

        :return: None
        :rtype: None
        """
        if self.subject_id:
            file_path = join(self.output_path, self.subject_id + "_")
            if self.session_id:
                file_path += self.session_id + "_"
            file_path += "blood.json"
        else:
            file_path = join(self.output_path, "blood.json")

        side_car_template = {
            "Time": {
                "Description": "Time in relation to time zero defined by the _pet.json",
                "Units": "s",
            },
            "whole_blood_radioactivity": {
                "Description": "Radioactivity in whole blood samples.",
                "Units": self.units,
            },
            "metabolite_parent_fraction": {
                "Description": "Parent fraction of the radiotracer",
                "Units": "arbitrary",
            },
        }

        if self.kwargs.get("MetaboliteMethod", None):
            side_car_template["MetaboliteMethod"] = (
                self.kwargs.get("MetaboliteMethod"),
            )
        elif self.kwargs.get("MetaboliteRecoveryCorrectionApplied", None):
            side_car_template["MetaboliteRecoveryCorrectionApplied"] = self.kwargs.get(
                "MetaboliteRecoveryCorrectionApplied"
            )
        elif self.kwargs.get("DispersionCorrected", None):
            side_car_template["DispersionCorrected"] = self.kwargs.get(
                "DispersionCorrected"
            )

        side_car_template["MetaboliteAvail"] = True

        if self.kwargs.get("MetaboliteMethod", None):
            side_car_template["MetaboliteMethod"] = self.kwargs.get("MetaboliteMethod")
        else:
            warnings.warn(
                "Parent fraction is available, but MetaboliteMethod is not specified, which is not BIDS "
                "compliant."
            )

        if self.kwargs.get("DispersionCorrected"):
            side_car_template["DispersionCorrected"] = self.kwargs.get(
                "DispersionCorrected"
            )
        else:
            warnings.warn(
                "Parent fraction is available, but there is no information if DispersionCorrected was"
                + "applied, which is not BIDS compliant"
            )

        if self.blood_series.get("plasma_activity", None) is type(pd.DataFrame):
            side_car_template["PlasmaAvail"] = True
            side_car_template["plasma_radioactivity"] = {
                "Description": "Radioactivity in plasma samples",
                "Units": self.units,
            }

        with open(file_path, "w") as out_json:
            json.dump(side_car_template, out_json, indent=4)

    def check_for_interpolated_data(self):
        # check to see if there may exist (emphasis on may) interpolated plasma values
        if type(self.blood_series.get("plasma_activity", None)) is not None:
            # extract parent fraction/metabolite fraction as series from dataframes
            metabolite_parent_fraction_series = pandas.Series(dtype="float64")
            for name, dataframe in self.blood_series.items():
                columns = dataframe.columns
                for entry in columns:
                    if "parent" in str.lower(entry) or "fraction" in str.lower(entry):
                        metabolite_parent_fraction_series = dataframe[str(entry)]
                        break
                if len(metabolite_parent_fraction_series) > 0:
                    break

            # check dataframes for plasma activity
            plasma_activity_series = pandas.Series(dtype="float64")
            for name, dataframe in self.blood_series.items():
                columns = dataframe.columns
                for entry in columns:
                    if "plasma_radioactivity" in entry:
                        plasma_activity_series = dataframe[str(entry)]
                        break
                if len(plasma_activity_series) > 0:
                    break

            if (
                len(plasma_activity_series) > 0
                and len(metabolite_parent_fraction_series) > 0
            ):
                if len(plasma_activity_series) == len(
                    metabolite_parent_fraction_series
                ):
                    logging.warning(
                        f"plasma_activity_series and metabolite_parent_fraction have same "
                        f"length {len(plasma_activity_series)}, plasma data may be interpolated, BIDS "
                        f"prefers raw data."
                    )

    def check_fraction_is_fraction(self):
        # collect parent fractions wherever they may exist
        fraction_series = []
        for name, dataframe in self.blood_series.items():
            for column in dataframe.columns:
                if "fraction" in str(column).lower():
                    fraction_series.append({name: dataframe[column]})

        for fraction_column in fraction_series:
            for name, values in fraction_column.items():
                for value in values:
                    logging.warning(
                        f"No, no my friend parent/metabolite fraction must be less than or equal to 1,"
                        f" found {value} in {name}."
                    )


def main():
    """
    Executes the PmodToBlood class using argparse

    :return: None
    """

    cli_args = cli()

    if cli_args.show_examples:
        print(example1)
    elif cli_args.whole_blood_path and cli_args.parent_fraction_path:
        pmod_to_blood = PmodToBlood(
            whole_blood_activity=cli_args.whole_blood_path,
            parent_fraction=cli_args.parent_fraction_path,
            plasma_activity=cli_args.plasma_activity_path,
            output_path=cli_args.output_path,
            output_json=cli_args.json,
            engine=cli_args.engine,
            kwargs=cli_args.kwargs,
        )
    else:
        raise Exception(
            f"--whole-blood-path (-w) and --parent-fraction-path (-p) are both required!"
        )


if __name__ == "__main__":
    main()



import typing
import os
import pathlib
import argparse
import logging
from json_maj.main import JsonMAJ

try:
    import helper_functions
    import pet_metadata as metadata
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions
    import pypet2bids.pet_metadata as metadata

# from pypet2bids.helper_functions import single_spreadsheet_reader, \
#    collect_bids_part, open_meta_data, load_pet_bids_requirements_json, ParseKwargs


def read_single_subject_spreadsheets(
    general_metadata_spreadsheet: pathlib.Path, **kwargs
) -> dict:
    """
    Reads in spreadsheet as formatted in PET2BIDS/spreadsheet_conversion/single_subject_sheet,
    extra arguments are supplied in key pair form via kwargs as is convention.

    :format:
    :param general_metadata_spreadsheet: path to a metadatspread sheet containing bids fields as columns
    with values below
    :type general_metadata_spreadsheet: file path
    :param kwargs: additional key pair arguments to pass on, these get applied generally
    just like the first spreadsheet. e.g. TimeZero=12:12:12, SpecificRadioactivity=3
    :type kwargs: string or dict
    :return: dictionary of subject data extracted from each spreadsheet along with any additional
    kwargs supplied
    :rtype: dict

    Anthony Galassi
    -----------------------------
    Copyright Open NeuroPET team
    """

    required_fields = metadata.PET_metadata

    subject_id = kwargs.get("subject_id", None)
    session_id = kwargs.get("session_id", None)

    if general_metadata_spreadsheet.is_file():
        general_metadata = helper_functions.single_spreadsheet_reader(
            general_metadata_spreadsheet
        )
        if not subject_id:
            subject_id = helper_functions.collect_bids_part(
                "sub", str(general_metadata_spreadsheet)
            )
            if subject_id:
                general_metadata["subject_id"] = subject_id
        else:
            general_metadata["subject_id"] = subject_id

        if not session_id:
            session_id = helper_functions.collect_bids_part(
                "ses", str(general_metadata_spreadsheet)
            )
            if session_id:
                general_metadata["session_id"] = session_id
        else:
            general_metadata["session_id"] = session_id

        column_set = list(general_metadata.keys())
        # assert required columns are in input spreadsheets
        for k in required_fields.keys():
            for field in required_fields[k]:
                field_exists = field in column_set
                if k == "mandatory" and not field_exists:
                    logging.warning(
                        f"Input spreadsheet {general_metadata_spreadsheet} is missing required "
                        f"column {field}"
                    )
                elif k == "recommended" and not field_exists:
                    logging.info(
                        f"Input spreadsheet(s) {general_metadata_spreadsheet} and is missing "
                        f"recommended column {field}"
                    )
                elif k == "optional":
                    logging.info(
                        f"Input spreadsheet(s) {general_metadata_spreadsheet} is missing "
                        f"optional column {field}"
                    )

        # check to see if there's a subject column in the multi subject data
        accepted_column_names = [
            "participant_id",
            "participant",
            "subject",
            "subject_id",
        ]
        columns = helper_functions.open_meta_data(
            metadata_path=general_metadata_spreadsheet
        ).columns
        found_column_names = []
        for acceptable in accepted_column_names:
            if acceptable in columns:
                found_column_names.append(acceptable)

        if len(found_column_names) > 1:
            error_message = (
                f"single_subject_spreadsheet: {general_metadata_spreadsheet} must contain only one column "
                f"of the following names: "
            )
            for name in accepted_column_names:
                error_message += name + " "
            error_message += f"\nContains these instead {found_column_names}."
            raise Exception(error_message)
        elif len(found_column_names) >= 0 and subject_id:
            if len(found_column_names) > 0:
                logging.warning(
                    f"Found subject id in filepath {general_metadata_spreadsheet} and column(s) "
                    f"{found_column_names}. Defaulting to the value {subject_id} found in the file path "
                    f"{general_metadata_spreadsheet}."
                )
        elif len(found_column_names) == 1 and not subject_id:
            subject_id_column = found_column_names[0]
            subject_id = general_metadata.pop(subject_id_column)
            general_metadata["subject_id"] = subject_id

        return general_metadata

    else:
        raise FileNotFoundError(general_metadata_spreadsheet)


def write_single_subject_spreadsheets(
    subject_metadata: dict,
    output_path: typing.Union[str, pathlib.Path],
    create_bids_tree: bool = False,
) -> str:
    """
    Writes out a dictionary of subjects to a series of json files, if files exist updates
    them with new values obtained from spreadsheets.
    :param subject_metadata: dictionary primary keys and all bids fields as values
    :type subject_metadata: dict
    :param output_path: path to write out files to, very much required
    :type output_path: str or pathlib.Path object
    :param create_bids_tree: boolean flag to create a bids tree, function will do it's best
        to create subject folders and session folders along w/ pet modality folder if this option is
        specified. Works on existing bids if session and subject id can be parsed from
        subject_metadata.
    :type create_bids_tree: bool
    :return: path to written json file
    :rtype: str
    """

    subject_id = subject_metadata.get("subject_id", None)
    if subject_id:
        subject_metadata.pop("subject_id")
    else:
        subject_id = helper_functions.collect_bids_part("sub", output_path)

    session_id = subject_metadata.get("session_id", None)
    if session_id:
        subject_metadata.pop("session_id")
    else:
        session_id = helper_functions.collect_bids_part("ses", output_path)

    if create_bids_tree:
        if subject_id not in output_path.parts:
            output_path = output_path / subject_id
        if session_id:
            json_out_file_name = subject_id + "_" + f"{session_id}_pet.json"
            if session_id not in output_path.parts:
                output_path = output_path / session_id
        else:
            json_out_file_name = subject_id + "_pet.json"

        if output_path.parts[-1] != "pet":
            output_path = output_path / "pet"

        output_path.mkdir(parents=True, exist_ok=True)

        JsonMAJ(
            json_path=os.path.join(output_path, json_out_file_name),
            update_values=subject_metadata,
        ).update()
    else:
        output_path.expanduser().mkdir(exist_ok=True, parents=True)
        json_out_path = os.path.join(output_path, f"{subject_id}")
        if session_id:
            json_out_path += f"_{session_id}_pet.json"
        else:
            json_out_path += "_pet.json"

        JsonMAJ(json_path=json_out_path, update_values=subject_metadata).update()

    return None


def cli():
    """
    Instantiates a command line interface with which allows reading and writing of
    a spreadsheet following the format specified in PET2BIDS/spreadsheet_conversion/single_subject_sheet/ to a json
    :return: a dictionary version of the subject sidecar json's that get written out
    :rtype: dict
    """
    parser = argparse.ArgumentParser(
        description="allows reading and writing of a spreadsheet following the format "
        "specified in PET2BIDS/spreadsheet_conversion/single_subject_sheet/ "
        "to a json"
    )
    parser.add_argument(
        "spreadsheet",
        type=pathlib.Path,
        help="Path to a spreadsheet with data applicable to multiple subjects",
    )
    parser.add_argument("--output-path", "-o", type=pathlib.Path)
    parser.add_argument("--bids-tree", "-b", action="store_true")
    parser.add_argument(
        "--kwargs", "-k", nargs="*", action=helper_functions.ParseKwargs, default={}
    )
    args = parser.parse_args()
    subject = read_single_subject_spreadsheets(
        general_metadata_spreadsheet=args.spreadsheet, **args.kwargs
    )
    if args.output_path:
        output_path = args.output_path.expanduser()
    else:
        output_path = pathlib.Path(os.getcwd())

    write_single_subject_spreadsheets(
        output_path=output_path.expanduser(),
        subject_metadata=subject,
        create_bids_tree=args.bids_tree,
    )

    return subject


if __name__ == "__main__":
    x = cli()



from json_maj.main import JsonMAJ
import pathlib
import numpy
import typing
import os
import argparse

from numpy.lib.recfunctions import require_fields

try:
    import pypet2bids.helper_functions as helper_functions
    import pypet2bids.pet_metadata as pet_metadata
except ModuleNotFoundError:
    import helper_functions
    import pet_metadata


def read_multi_subject_spreadsheets(
    general_metadata_spreadsheet: pathlib.Path,
    multiple_subject_spreadsheet: pathlib.Path,
    **kwargs,
) -> dict:
    """
    Reads in two spreadsheets as formatted in PET2BIDS/spreadsheet_conversion/many_subject_sheets,
    generic (scanner or subject independent data is supplied via the first argument and subject
    specific data is supplied via the second argument.

    :param general_metadata_spreadsheet: path to a metadata spreadsheet containing bids fields as columns
        with values below
    :type general_metadata_spreadsheet: file path
    :param multiple_subject_spreadsheet: path to multi subject spreadsheet containing a subject id, participant id,
        subject, or participant column consisting of paths to subject folders/files.
    :type multiple_subject_spreadsheet: file path
    :param kwargs: additional key pair arguments to pass on, these get applied generally
        just like the first spreadsheet. e.g. TimeZero=12:12:12, SpecificRadioactivity=3
    :type kwargs: string or dict
    :return: dictionary of subject data extracted from each spreadsheet along with any additional
        kwargs supplied
    :rtype: dict

    Anthony Galassi
    Copyright Open NeuroPET team

    """

    required_fields = pet_metadata.PET_metadata

    if (
        general_metadata_spreadsheet.is_file()
        and multiple_subject_spreadsheet.is_file()
    ):
        general_metadata = helper_functions.single_spreadsheet_reader(
            general_metadata_spreadsheet
        )
        multiple_subject_metadata = helper_functions.open_meta_data(
            multiple_subject_spreadsheet
        )

        column_set = set(
            list(general_metadata.keys()) + list(multiple_subject_metadata.columns)
        )

        column_set = list(column_set)

        # assert required columns are in input spreadsheets
        for k in required_fields.keys():
            for field in required_fields[k]:
                field_exists = field in column_set
                if k == "mandatory" and not field_exists:
                    helper_functions.logging.warning(
                        f"Input spreadsheet(s) {general_metadata_spreadsheet} and "
                        f"{multiple_subject_spreadsheet} are missing required column {field}"
                    )
                elif k == "recommended" and not field_exists:
                    helper_functions.logging.info(
                        f"Input spreadsheet(s) {general_metadata_spreadsheet} and "
                        f"{multiple_subject_spreadsheet} are missing recommended column {field}"
                    )
                elif k == "optional":
                    helper_functions.logging.info(
                        f"Input spreadsheet(s) {general_metadata_spreadsheet} and "
                        f"{multiple_subject_spreadsheet} are missing optional column {field}"
                    )

        # check to see if there's a subject column in the multi subject data
        accepted_column_names = [
            "participant_id",
            "participant",
            "subject",
            "subject_id",
        ]
        found_column_names = []
        for acceptable in accepted_column_names:
            if acceptable in multiple_subject_metadata.columns:
                found_column_names.append(acceptable)

        if len(found_column_names) != 1:
            error_message = (
                f"multi-subject_spreadsheet: {multiple_subject_spreadsheet} must contain only one column "
                f"of the following names: "
            )
            for name in accepted_column_names:
                error_message += name + " "
            error_message += f"\nContains these instead {found_column_names}."
            raise Exception(error_message)
        else:
            subject_column = found_column_names[0]

        # collect all subject id's
        subject_metadata = {}
        for subject in multiple_subject_metadata.get(subject_column, None):
            subject_row = helper_functions.get_coordinates_containing(
                subject, multiple_subject_metadata, single=True
            )[0]
            subject_id = helper_functions.collect_bids_part("sub", subject)
            session_id = helper_functions.collect_bids_part("ses", subject)
            if subject_id:
                subject_metadata[subject_id] = general_metadata
                if session_id:
                    subject_metadata[subject_id]["session_id"] = session_id
                if kwargs:
                    subject_metadata[subject_id].update(**kwargs)

                subject_data_from_row = helper_functions.transform_row_to_dict(
                    subject_row, multiple_subject_metadata
                )
                for k, v in subject_data_from_row.items():
                    if v and k != subject_column and v is not numpy.nan:
                        subject_metadata[subject_id][k] = v

        return subject_metadata

    else:
        missing_files = [
            missing
            for missing in (general_metadata_spreadsheet, multiple_subject_spreadsheet)
            if not missing.is_file()
        ]
        error_message = f"Missing "
        for m in missing_files:
            error_message += f"{m=}".split("=")[0]
            error_message += f" {m}"
        raise Exception(error_message)


def write_multi_subject_spreadsheets(
    subjects: dict,
    output_path: typing.Union[str, pathlib.Path],
    create_bids_tree: bool = False,
) -> None:
    """
    Writes out a dictionary of subjects to a series of json files, if files exist updates
    them with new values obtained from spreadsheets.

    :param subjects: subject dictionary with subject id as primary keys and all bids fields as values
    :type subjects: dict
    :param output_path: path to write out files to, very much required
    :type output_path: str or pathlib.Path object
    :param create_bids_tree: boolean flag to create a bids tree, function will do it's best
        to create subject folders and session folders along w/ pet modality folder if this option is
        specified. Works on existing bids trees so long as session and subject id can be parsed from
        multi subject input sheet.
    :type create_bids_tree: bool
    :return: None
    :rtype: None
    """
    if create_bids_tree:
        for subject, fields in subjects.items():
            json_out_path = os.path.join(output_path, f"{subject}")
            if fields.get("session_id", None):
                json_out_file_name = (
                    subject + "_" + f"{fields.get('session_id')}_pet.json"
                )
                json_out_path = os.path.join(
                    json_out_path, fields.get("session_id"), "pet"
                )
            else:
                json_out_file_name = subject + "_pet.json"
                json_out_path = os.path.join(json_out_path, "pet")
            pathlib.Path(json_out_path).mkdir(parents=True, exist_ok=True)
            json_out_file_name = os.path.join(json_out_path, json_out_file_name)
            json_out = JsonMAJ(
                json_path=json_out_file_name, update_values=fields
            ).update()
    else:
        for subject, fields in subjects.items():
            json_out_path = os.path.join(output_path, f"{subject}")
            if fields.get("session_id", None):
                json_out_path += f"{fields.get('session_id')}_pet.json"
            else:
                json_out_path += "_pet.json"

            json_out = JsonMAJ(json_path=json_out_path, update_values=fields).update()


def cli():
    """
    Instantiates a command line interface with which allows reading and writing of
    2 spreadsheets following the format specified in PET2BIDS/spreadsheet_conversion/many_subjects_sheet
    :return: a dictionary version of the subject sidecar json's that get written out
    :rtype: dict
    """
    parser = argparse.ArgumentParser(
        description="allows reading and writing of 2 spreadsheets following the format "
        "specified in PET2BIDS/spreadsheet_conversion/many_subjects_sheet"
    )
    parser.add_argument(
        "--general-spreadsheet",
        "-g",
        type=pathlib.Path,
        help="Path to a spreadsheet with data applicable to multiple subjects",
    )
    parser.add_argument(
        "--many-subjects-spreadsheet",
        "-m",
        type=pathlib.Path,
        help="Path to spreadsheet containing multiple subjects",
    )
    parser.add_argument("--output-path", "-o", type=pathlib.Path)
    parser.add_argument("--bids-tree", "-b", action="store_true")
    args = parser.parse_args()
    subjects = read_multi_subject_spreadsheets(
        general_metadata_spreadsheet=args.general_spreadsheet,
        multiple_subject_spreadsheet=args.many_subjects_spreadsheet,
    )
    if args.output_path:
        output_path = args.output_path
    else:
        output_path = os.getcwd()

    write_multi_subject_spreadsheets(
        output_path=output_path, subjects=subjects, create_bids_tree=args.bids_tree
    )

    return subjects


if __name__ == "__main__":
    x = cli()



"""
This module contains methods used to read ecat files (\*.v), the primary method pulled/imported from this module
is read_ecat which returns the contents of a singular ecat file divided into three parts:

- main header
- subheaders
- pixel data

The other functions primarily operate bitwise, so it's best not to use them outside of this module, unless you are a
glutton or twisted in that sort of way.

:Authors: Anthony Galassi
:Copyright: Open NeuroPET team

"""

import json
import os.path
import struct
from os import path
from os.path import join
import pathlib
import re
import numpy
from pypet2bids.helper_functions import decompress, first_middle_last_frames_to_text

parent_dir = pathlib.Path(__file__).parent.resolve()
code_dir = parent_dir.parent
data_dir = code_dir.parent

# debug variable
# save steps for debugging for more info see ecat_testing/README.md
ecat_save_steps = os.environ.get("ECAT_SAVE_STEPS", 0)
if ecat_save_steps == "1":
    # check to see if the code directory is available, if it's not create it and
    # the steps dir to save outputs created if ecat_save_steps is set to 1
    steps_dir = code_dir.parent / "ecat_testing" / "steps"
    if not steps_dir.is_dir():
        os.makedirs(code_dir.parent / "ecat_testing" / "steps", exist_ok=True)


# collect ecat header maps, this program will not work without these as ECAT data varies in the byte location of its
# data depending on the version of ECAT it was formatted with.
try:
    with open(join(parent_dir, "ecat_headers.json"), "r") as infile:
        ecat_header_maps = json.load(infile)
except FileNotFoundError:
    raise Exception(
        "Unable to load header definitions and map from ecat_headers.json. Aborting."
    )


# noinspection PyShadowingNames
def get_ecat_bytes(path_to_ecat: str):
    """
    Opens an ecat file and reads the entry file into memory to return a bytes object
    not terribly memory efficient for large or parallel reading of ecat files.

    :param path_to_ecat: path to an ecat file, however will literally open any file an read it
        in as bytes.
    :return: a bytes object
    """
    # check if file exists
    if path.isfile(path_to_ecat):
        with open(path_to_ecat, "rb") as infile:
            ecat_bytes = infile.read()
    else:
        raise Exception(f"No such file found at {path_to_ecat}")

    return ecat_bytes


def read_bytes(path_to_bytes: str, byte_start: int, byte_stop: int = -1):
    """
    Open a file at path to bytes and reads in the information byte by byte.

    :param path_to_bytes: Path to file to read
    :param byte_start: Position to place the seek head at before reading bytes
    :param byte_stop: Position to stop reading bytes at
    :return: the bytes located at the position sought when invoking this function.
    """
    if not os.path.isfile(path_to_bytes):
        raise Exception(f"{path_to_bytes} is not a valid file.")

    # open that file
    bytes_to_read = open(path_to_bytes, "rb")

    # move to start byte
    bytes_to_read.seek(byte_start, 0)

    # read a section of bytes from bytestart to byte stop
    byte_width = byte_stop
    sought_bytes = bytes_to_read.read(byte_width)

    bytes_to_read.close()
    return sought_bytes


def collect_specific_bytes(
    bytes_object: bytes, start_position: int = 0, width: int = 0
):
    """
    Collects specific bytes within a bytes object.

    :param bytes_object: an opened bytes object
    :param start_position: the position to start to read at
    :param width: how far to read from the start position
        2 -> end of file/object
    :return: the bytes starting at position
    """
    # navigate to byte position
    content = bytes_object[start_position : start_position + width]
    return {"content": content, "new_position": start_position + width}


def get_buffer_size(data_type: str, variable_name: str):
    """
    Determine the byte width of a variable as defined in the ecat_headers.json
    such that Fill(6) will return 6

    :param data_type:
    :param variable_name:
    :return: the number of bytes to expand a buffer to
    """
    first_split = variable_name.split("(")
    if len(first_split) == 2:
        fill_scalar = int(first_split[1][:-1])
    else:
        fill_scalar = 1
    scalar = int(re.findall(r"\d+", data_type)[0]) * fill_scalar

    return scalar


def get_header_data(
    header_data_map: dict, ecat_file: str = "", byte_offset: int = 0, clean=True
):
    """
    ECAT header data is contained in json files translated from ECAT documentation provided via the Turku PET Inst.
    For machine and human readability the original Siemens PDF/Scanned ECAT file documentation has been rewritten into
    json to provide easy access to byte position, byte width, variable name, and byte data type. Json's including this
    header information are sub-divided into ECAT versions with sub-dictionaries corresponding to each imaging type as
    mentioned in the original ECAT spec.

    :param header_data_map: schema for reading in header data, this is stored in dictionary that is read in from a
           json file.
    :param ecat_file: path to an ecat file, file is opened and byte blocks containing header data are then extracted,
           read, and cleaned into python float, int, or list data types
    :param byte_offset: off set from the head of the file to read from
    :param clean: Whether to remove byte padding or not, if not provided strings will include padding w/ \0x bytes and
           lists/arrays of data will be returned as tuples instead of lists. Uncleaned data will be of format b''.
    :return: a dictionary with variable names of header fields as keys and cleaned/uncleaned header fields as values
    """
    header = {}

    for value in header_data_map:
        byte_position, variable_name, struct_fmt = (
            value["byte"],
            value["variable_name"],
            ">" + value["struct"],
        )
        byte_width = struct.calcsize(struct_fmt)
        relative_byte_position = byte_position + byte_offset
        raw_bytes = read_bytes(ecat_file, relative_byte_position, byte_width)
        header[variable_name] = struct.unpack(struct_fmt, raw_bytes)
        if clean and "fill" not in variable_name.lower():
            header[variable_name] = filter_bytes(header[variable_name], struct_fmt)
        if "comment" in value and "msec" in value["comment"]:
            # for entries that are in msec, convert to sec for PET BIDS json
            header[variable_name] /= 1000
        read_head_position = relative_byte_position + byte_width

    return header, read_head_position


def filter_bytes(unfiltered: bytes, struct_fmt: str):
    """
    Cleans up byte strings and bytes types into python int, float, string, and list data types, additionally
    struct.unpack returns a tuple object even if only a single value is available, this function determines when to
    return a single value or a list of values based on the contents of the unfiltered object.

    :param unfiltered: a raw bytes type object
    :param struct_fmt: the c struct type of the object
    :return: a cleaned python int, float, string, or list
    """
    if len(unfiltered) == 1:
        unfiltered = unfiltered[0]
    elif len(unfiltered) > 1:
        unfiltered = list(unfiltered)
    if "s" in struct_fmt:
        filtered = str(bytes(filter(None, unfiltered)), "UTF-8")
    else:
        filtered = unfiltered
    return filtered


def get_directory_data(byte_block: bytes, ecat_file: str, return_raw: bool = False):
    """
    Collects the directory data within an ECAT file. The directory data refers to the 512 byte table that describes the
    byte location of each frame, subheader, number of frames, and additional directory tables within the file.

    :param byte_block: A block of file bytes to convert into a 2 dimensional numpy array.
    :param ecat_file: the path to the ecat file
    :param return_raw: return the directory tables as extracted, if left False this will return the directory tables
           combined into a single table. The single table is all that is needed in order to read information in from an
           ECAT.
    :return: Individual tables corresponding to up to 31 frames each or a combined directory table consisting of no
             more columns than than are number of frames in the image/PET scan.
    """
    directory = None  # used to keep track of state in the event of a directory spanning more than one 512 byte block
    raw = []
    while True:
        # The exit conditions for this loop are below
        # if [4,1] of the directory is 0 break as there are 31 or less frames in this 512 byte buffer
        # if [2,1] of the directory is 2 break ????, up for interpretation as to the exact meaning but,
        # observed to signal the end of an additional 512 byte block/buffer when the number of frames
        # exceeds 31

        read_byte_array = numpy.frombuffer(
            byte_block, dtype=numpy.dtype(">i4"), count=-1
        )
        # reshape 1d array into 2d, a 4 row by 32 column table is expected
        reshaped = numpy.transpose(numpy.reshape(read_byte_array, (-1, 4)))

        raw.append(reshaped)

        # chop off columns after 32, rows after 32 appear to be noise
        reshaped = reshaped[:, 0 : read_byte_array[3] + 1]
        # get directory size/number of frames in dir from 1st column 4th row of the array in the buffer
        directory_size = reshaped[3, 0]
        if directory_size == 0:
            break
        # on the first pass do this
        if directory is None:
            directory = reshaped[:, 1 : directory_size + 1]
        else:
            directory = numpy.append(
                directory, reshaped[:, 1 : directory_size + 1], axis=1
            )
        # determine if this is the last directory by examining the 2nd row of the first column of the buffer
        next_directory_position = reshaped[1, 0]
        if next_directory_position == 2:
            break
        # looks like there is more directory to read, collect some more bytes
        byte_block = read_bytes(
            path_to_bytes=ecat_file,
            byte_start=(next_directory_position - 1) * 512,
            byte_stop=512,
        )

    # sort the directory contents as they're sometimes out of order
    sorted_directory = directory[:, directory[0].argsort()]

    if return_raw:
        return raw
    else:
        return sorted_directory


def read_ecat(
    ecat_file: str, calibrated: bool = False, collect_pixel_data: bool = True
):
    """
    Reads in an ecat file and collects the main header data, subheader data, and imagining data.

    :param ecat_file: path to an ecat file, does not handle compression currently
    :param calibrated: if True, will scale the raw imaging data by the SCALE_FACTOR in the subheader and
    :param collect_pixel_data: By default collects the entire ecat, can be passed false to only return headers
        CALIBRATION_FACTOR in the main header

    :return: main_header, a list of subheaders for each frame, the imagining data from the subheaders

    """
    if ".gz" in ecat_file:
        ecat_file = decompress(ecat_file)

    # try to determine what type of ecat this is
    possible_ecat_headers = {}
    for entry, dictionary in ecat_header_maps["ecat_headers"].items():
        possible_ecat_headers[entry] = dictionary["mainheader"]

    # set byte order
    byte_order = ">"

    confirmed_version = None
    for version, dictionary in possible_ecat_headers.items():
        try:
            possible_header, _ = get_header_data(dictionary, ecat_file)
            if version == str(possible_header["SW_VERSION"]):
                confirmed_version = version
                break
        except UnicodeDecodeError:
            continue

    if not confirmed_version:
        raise Exception(
            f"Unable to determine ECAT File Type from these types {possible_ecat_headers.keys()}"
        )

    ecat_main_header = ecat_header_maps["ecat_headers"][confirmed_version]["mainheader"]

    main_header, read_to = get_header_data(ecat_main_header, ecat_file)

    if ecat_save_steps == "1":
        with open(steps_dir / "1_read_mh_ecat_python.json", "w") as outfile:
            json.dump(main_header, outfile)
    """
    Some notes about the file directory/sorted directory:
    
    Comments referencing matrix/table indexing may vary by +-1 in relation to code written.
    Python is 0 indexed by default so it can be taken as truth w/ relation to element location.
    Deviation from this convention are intended to clarify what is happening to a human reader
    although we are aware that this most likely has the opposite effect. 

    The first or 0th column of the file directory correspond to the nature of the directory itself:
    row 0: ??? No idea, some integer
    row 1: Byte position of this table/directory
    row 2: not sure in testing it seems to be 0 most times..
    row 3: The number of frames/additional columns in the file. If the number of columns of this array
    is n, it would contain n-1 frames. 

    The values in sorted_directory correspond to the following for all columns except the first column
    row 0: Not sure, but we sort on this, perhaps it's the frame start time
    row 1: the start byte block position of the frame data
    row 2: end byte block position of the frame data
    row 3: ??? Number of frames contained in w/ in the byte blocks between row 1 and 2?
    """

    # Collecting First Part of File Directory/Index this Directory lies directly after the main header byte block
    next_block = read_bytes(
        path_to_bytes=ecat_file, byte_start=read_to, byte_stop=read_to + 512
    )

    directory = get_directory_data(next_block, ecat_file)

    # determine subheader type by checking main header
    subheader_type_number = main_header["FILE_TYPE"]

    """
    ECAT 7.2 Only
    Subheader types correspond to these enumerated types as defined below:
    00 = unknown, 
    01 = Sinogram, 
    02 = Image - 16, 
    03 = Attenuation Correction, 
    04 = Normalization, 
    05 = PolarMap, 
    06 = Volume 8, 
    07 = Volume 16, 
    08 = Projection 8, 
    09 = Projection 16, 
    10 = Image 8, 
    11 = 3D Sinogram 16, 
    12 = 3D Sinogram 8, 
    13 = 3D Normalization, 
    14 = 3D Sinogram Fit)

    Presently, only types 03, 05, 07, 11, and 13 correspond to known subheader types for 72. If the
    value in FILE_TYPE is outside of this range the subheaders will not be read and this will
    raise an exception.
    
    ECAT 7.3 Only
    00 = unknown
    01 = unknown
    02 = unknown
    03 = Attenuation Correction
    04 = unknown
    05 = unknown
    06 = unknown
    07 = Volume 16
    08 = unknown
    09 = unknown
    10 = unknown
    11 = 3D Sinogram 16 
    """

    # collect the bytes map file for the designated subheader, note some are not supported.
    subheader_map = ecat_header_maps["ecat_headers"][confirmed_version][
        str(subheader_type_number)
    ]

    if not subheader_map:
        raise Exception(f"Unsupported data type: {subheader_type_number}")

    # collect subheaders and pixel data
    subheaders, data = [], []
    for i in range(len(directory.T)):
        frame_number = i + 1
        if collect_pixel_data:
            print(f"Reading subheader from frame {frame_number}")

        # collect frame info/column
        frame_info = directory[:, i]
        frame_start = frame_info[1]
        frame_stop = frame_info[2]

        frame_start_byte_position = 512 * (frame_start - 1)  # sure why not
        # read subheader
        subheader, byte_position = get_header_data(
            subheader_map, ecat_file, byte_offset=frame_start_byte_position
        )

        if collect_pixel_data:
            # collect pixel data from file
            pixel_data = read_bytes(
                path_to_bytes=ecat_file,
                byte_start=512 * frame_start,
                byte_stop=512 * frame_stop,
            )

            # calculate size of matrix for pixel data, may vary depending on image type (polar, 3d, etc.)
            if subheader_type_number == 7:
                image_size = [
                    subheader["X_DIMENSION"],
                    subheader["Y_DIMENSION"],
                    subheader["Z_DIMENSION"],
                ]
                # check subheader for pixel datatype
                dt_val = subheader["DATA_TYPE"]
                if dt_val == 5:
                    formatting = ">f4"
                    pixel_data_type = numpy.dtype(formatting)
                elif dt_val == 6:
                    # >H is unsigned short e.g. >u2, reverting to int16 e.g. i2 to align with commit 9beee53
                    pixel_data_type = ">i2"
                else:
                    raise ValueError(
                        f"Unable to determine pixel data type from value: {dt_val} extracted from {subheader}"
                    )
                # read it into a one dimensional matrix
                pixel_data_matrix_3d = numpy.frombuffer(
                    pixel_data,
                    dtype=pixel_data_type,
                    count=image_size[0] * image_size[1] * image_size[2],
                ).reshape(*image_size, order="F")
                # pixel_data_matrix_3d.newbyteorder(byte_order)

                if ecat_save_steps == "1":
                    with open(
                        steps_dir / f"4_determine_data_type_python.json", "w"
                    ) as outfile:
                        json.dump(
                            {"DATA_TYPE": dt_val, "formatting": pixel_data_type},
                            outfile,
                        )

            else:
                raise Exception(
                    f"Unable to determine frame image size, unsupported image type {subheader_type_number}"
                )

            # we assume the user doesn't want to do multiplication to adjust for calibration here
            if calibrated:
                calibration_factor = (
                    subheader["SCALE_FACTOR"] * main_header["ECAT_CALIBRATION_FACTOR"]
                )
                calibrated_pixel_data_matrix_3d = (
                    calibration_factor * pixel_data_matrix_3d
                )
                data.append(calibrated_pixel_data_matrix_3d)
            else:
                data.append(pixel_data_matrix_3d)
        else:
            data = None

        subheaders.append(subheader)

        if ecat_save_steps == "1":
            with open(steps_dir / f"2_read_sh_ecat_python.json", "w") as outfile:
                json.dump({"subheaders": subheaders}, outfile)

    if collect_pixel_data:
        # return 4d array instead of list of 3d arrays
        pixel_data_matrix_4d = numpy.zeros(
            tuple(image_size + [len(data)]), dtype=numpy.dtype(pixel_data_type)
        )
        for index, frame in enumerate(data):
            pixel_data_matrix_4d[:, :, :, index] = frame

        if ecat_save_steps == "1":

            # write out the endianness and datatype of the pixel data matrix
            step_3_dict = {
                "datatype": pixel_data_matrix_4d.dtype.name,
                "endianness": pixel_data_matrix_4d.dtype.byteorder,
            }

            with open(steps_dir / f"3_determine_data_type_python.json", "w") as outfile:
                json.dump(step_3_dict, outfile)

            # record out step 4
            first_middle_last_frames_to_text(
                four_d_array_like_object=pixel_data_matrix_4d,
                output_folder=steps_dir,
                step_name="4_read_img_ecat_python",
            )

            # record out step 5
            if calibrated:
                first_middle_last_frames_to_text(
                    four_d_array_like_object=calibrated_pixel_data_matrix_3d,
                    output_folder=steps_dir,
                    step_name="5_scale_img_ecat_python",
                )

    else:
        pixel_data_matrix_4d = None

    return main_header, subheaders, pixel_data_matrix_4d



"""
This module contains a collection of general functions used across this library, that is to say if a method ends up here
it is because it is useful in more than one context.

Some of the modules in this library that depend on this module are:

- :meth:`pypet2bids.convert_pmod_to_blood`
- :meth:`pypet2bids.dcm2niix4pet`
- :meth:`pyet2bids.read_ecat`
- :meth:`pypet2bids.ecat`
- :meth:`pypet2bids.ecat2nii`
- :meth:`pypet2bids.multiple_spreadsheets`

| *Authors: Anthony Galassi*
| *Copyright OpenNeuroPET team*
"""

import os
import gzip
import re
import shutil
import typing
import json
import hashlib
import warnings
import logging
import dotenv
import ast
import sys
import subprocess

import numpy
import pandas
import toml
import pathlib
from pandas import read_csv, read_excel, Series
from pathlib import Path
import importlib
import argparse
from typing import Union
from platform import system
import importlib

try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

try:
    import pet_metadata as metadata
except ImportError:
    import pypet2bids.pet_metadata as metadata

# load bids schema
schema = metadata.schema
pet_metadata = metadata.PET_metadata
# putting these paths here as they are reused in dcm2niix4pet.py, update_json_pet_file.py, and ecat.py
module_folder = Path(__file__).parent.resolve()
python_folder = module_folder.parent
pet2bids_folder = python_folder.parent

loggers = {}


def logger(name):
    global loggers
    if loggers.get(name):
        return loggers.get(name)
    else:
        # create logger with for pypet2bids
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)

        # create console handler with a higher log level
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(logging.DEBUG)

        ch.setFormatter(CustomFormatter())

        logger.addHandler(ch)
        # this stops the logger from repeating itself in the outputs, it's most likely set up incorrectly, but this
        # works great to get the desired effect
        logger.propagate = False

        loggers[name] = logger
        return logger


def flatten_series(series):
    """
    This function retrieves either a list or a single value from a pandas series object thus converting a complex
    data type to a simple datatype or list of simple types. If the length of the series is one or less this returns that
    single value, else this object returns all values within the series that are not Null/nan in the form of a list
    :param series: input series of type pandas.Series object, typically extracted as a column/row from a
    pandas.Dataframe object
    :return: a simplified single value or list of values
    """
    simplified_series_object = series.dropna().to_list()
    if len(simplified_series_object) > 1:
        pass
    elif len(simplified_series_object) == 1:
        simplified_series_object = simplified_series_object[0]
    else:
        raise Exception(f"Invalid Series: {series}")
    return simplified_series_object


def collect_spreadsheets(folder_path: pathlib.Path):
    spreadsheet_files = []
    if folder_path.is_file():
        spreadsheet_files.append(folder_path)
    else:
        all_files = [
            folder_path / pathlib.Path(file)
            for file in os.listdir(folder_path)
            if os.path.isfile(os.path.join(folder_path, file))
        ]
        for file in all_files:
            if (
                file.suffix == ".xlsx"
                or file.suffix == ".csv"
                or file.suffix == ".xls"
                or file.suffix == ".tsv"
            ):
                spreadsheet_files.append(file)
    return spreadsheet_files


def single_spreadsheet_reader(
    path_to_spreadsheet: Union[str, pathlib.Path],
    pet2bids_metadata: dict = metadata.PET_metadata,
    dicom_metadata={},
    **kwargs,
) -> dict:

    spreadsheet_metadata = {}
    metadata_fields = pet2bids_metadata

    if type(path_to_spreadsheet) is str:
        path_to_spreadsheet = pathlib.Path(path_to_spreadsheet)

    if path_to_spreadsheet.is_file():
        pass
    else:
        raise FileNotFoundError(f"{path_to_spreadsheet} does not exist.")

    spreadsheet_dataframe = open_meta_data(path_to_spreadsheet)

    log = logging.getLogger("pypet2bids")

    # collect mandatory fields
    for field_level in metadata_fields.keys():
        for field in metadata_fields[field_level]:
            series = spreadsheet_dataframe.get(field, Series(dtype=numpy.float64))
            if not series.empty:
                spreadsheet_metadata[field] = flatten_series(series)
            elif (
                series.empty
                and field_level == "mandatory"
                and not dicom_metadata.get(field, None)
                and field not in kwargs
            ):
                log.warning(
                    f"{field} not found in metadata spreadsheet: {path_to_spreadsheet}, {field} is required by BIDS"
                )

    # lastly apply any kwargs to the metadata
    spreadsheet_metadata.update(**kwargs)

    # more lastly, check to see if values are of the correct datatype (e.g. string, number, boolean)
    for field, value in spreadsheet_metadata.items():
        # check schema for field
        field_schema_properties = schema["objects"]["metadata"].get(field, None)
        if field_schema_properties:
            # check to see if value is of the correct type
            if field_schema_properties.get("type") == "number":
                if not is_numeric(str(value)):
                    log.warning(f"{field} is not numeric, it's value is {value}")
            if field_schema_properties.get("type") == "boolean":
                if type(value) is not bool:
                    try:
                        check_bool = int(value) / 1
                        if check_bool == 0 or check_bool == 1:
                            spreadsheet_metadata[field] = bool(value)
                        else:
                            log.warning(
                                f"{field} is not boolean, it's value is {value}"
                            )
                    except ValueError:
                        pass
            elif field_schema_properties.get("type") == "string":
                if type(value) is not str:
                    log.warning(f"{field} is not string, it's value is {value}")
            else:
                pass
    return spreadsheet_metadata


def compress(file_like_object, output_path: str = None):
    """
    Compresses a file using gzip.

    :param file_like_object: a file path to an uncompressed file
    :param output_path: an output path to compress the file to, if omitted simply appends .gz to
        file_like_object path
    :return: output_path on successful completion of compression
    """
    file_like_object = pathlib.Path(file_like_object)

    if file_like_object.exists() and not output_path:
        old_suffix = file_like_object.suffix
        if ".gz" not in old_suffix:
            output_path = file_like_object.with_suffix(old_suffix + ".gz")
        else:
            output_path = file_like_object

    elif not os.path.isfile(file_like_object):
        raise Exception(f"{file_like_object} is not a valid file to compress.")
    else:
        pass

    with open(file_like_object, "rb") as infile:
        input_data = infile.read()

    output = gzip.GzipFile(output_path, "wb")
    output.write(input_data)
    output.close()

    if output_path.exists():
        file_like_object.unlink(missing_ok=True)

    return output_path


def decompress(file_like_object, output_path: str = None):
    """
    Decompresses a gzip file.

    :param file_like_object: a compressed gzip file
    :param output_path: optional output path, if not supplied this function simply trims '.gz' off of
        the input file and writes to that amended path
    :return: output_path on successful decompression
    """
    if not output_path and ".gz" in file_like_object:
        output_path = re.sub(".gz", "", file_like_object)

    compressed_file = gzip.GzipFile(file_like_object)
    compressed_input = compressed_file.read()
    compressed_file.close()

    with open(output_path, "wb") as outfile:
        outfile.write(compressed_input)
    return output_path


def load_vars_from_config(
    path_to_config: str = pathlib.Path.home() / ".pet2bidsconfig",
):
    """
    Loads values from a .env file given a path to said .env file.

    :param path_to_config: path to .env file
    :return: a dictionary containing the values stored in .env
    """
    if os.path.isfile(path_to_config):
        parameters = dotenv.main.dotenv_values(path_to_config)
    else:
        log = logger("pypet2bids")
        log.warning(f"Unable to locate {path_to_config}, returning empty dictionary.")
        parameters = {}
        # raise FileNotFoundError(path_to_config)

    for parameter, value in parameters.items():
        try:
            parameters[parameter] = very_tolerant_literal_eval(value)
        except (ValueError, SyntaxError):
            parameters[parameter] = str(value)

    return parameters


def get_version():
    """
    Gets the version of this software
    :return: version number
    """
    # this scripts directory path
    scripts_dir = pathlib.Path(os.path.dirname(__file__))
    try:
        version = metadata.version("pypet2bids")
    except Exception:
        version = None

    if not version:
        tomlfile = {}
        try:
            # if this is bundled as a package look next to this file for the pyproject.toml
            toml_path = os.path.join(scripts_dir, "pyproject.toml")
            with open(toml_path, "r") as infile:
                tomlfile = toml.load(infile)
        except FileNotFoundError:
            # when in development the toml file with the version is 2 directories above (e.g. where it should actually live)
            try:
                toml_dir = scripts_dir.parent
                toml_path = os.path.join(toml_dir, "pyproject.toml")
                with open(toml_path, "r") as infile:
                    tomlfile = toml.load(infile)
            except FileNotFoundError:
                pass

        attrs = tomlfile.get("tool", {})
        poetry = attrs.get("poetry", {})
        version = poetry.get("version", "")

    return version


def is_numeric(check_this_object: str) -> bool:
    try:
        converted_to_num = ast.literal_eval(check_this_object)
        numeric_types = [int, float, pandas.NA]
        if type(converted_to_num) in numeric_types:
            return True
    except (ValueError, TypeError):
        return False


class ParseKwargs(argparse.Action):
    """
    Class that is used to extract key pair arguments passed to an argparse.ArgumentParser object via the command line.
    Accepts key value pairs in the form of 'key=value' and then passes these arguments onto the arg parser as kwargs.
    This class is used during the construction of the ArgumentParser class via the add_argument method. e.g.:\n
    `ArgumentParser.add_argument('--kwargs', '-k', nargs='*', action=helper_functions.ParseKwargs, default={})`
    """

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        for value in values:
            try:
                key, value = value.split("=")
                getattr(namespace, self.dest)[key] = very_tolerant_literal_eval(value)
            except ValueError:
                raise Exception(f"Unable to unpack {value}")


def very_tolerant_literal_eval(value):
    """
    Evaluates a string or string like input into a python datatype. Provides a lazy way to extract True from 'true',
    None from 'none', [0] from '[0'], etc. etc.
    :param value: the value you wish to convert to a python type
    :type value: string like, could be anything that can be evaluated as valid python
    :return: the value converted into a python object
    :rtype: depends on what ast.literal_eval determines the object to be
    """
    try:
        value = ast.literal_eval(value)
    except (SyntaxError, ValueError):
        if str(value).lower() == "none":
            value = None
        elif str(value).lower() == "true":
            value = True
        elif str(value).lower() == "false":
            value = False
        elif str(value)[0] == "[" and str(value)[-1] == "]":
            array_contents = str(value).replace("[", "").replace("]", "")
            array_contents = array_contents.split(",")
            array_contents = [str.strip(content) for content in array_contents]
            # evaluate array contents one by one
            value = [very_tolerant_literal_eval(v) for v in array_contents]
        else:
            value = str(value)
    return value


def open_meta_data(
    metadata_path: Union[str, pathlib.Path], separator=None
) -> pandas.DataFrame:
    """
    Opens a text metadata file with the pandas method most appropriate for doing so based on the metadata
    file's extension.
    :param metadata_path: Path or pathlike object/string to a spreadsheet file
    :type metadata_path: Path or str
    :param separator: Optional separator argument, used to try and parse tricky spreadsheets. e.g. ',' '\t', ' '
    :type separator: str
    :return: a pandas dataframe representation of the spreadsheet/metadatafile
    """
    log = logger("pypet2bids")
    if type(metadata_path) is str:
        metadata_path = pathlib.Path(metadata_path)

    separators_present = []
    if metadata_path.exists():
        pass
    else:
        raise FileExistsError(metadata_path)

    # collect suffix from metadata and use the appropriate pandas method to read the data
    extension = metadata_path.suffix

    methods = {"excel": read_excel, "csv": read_csv, "tsv": read_csv, "txt": read_csv}

    if "xls" in extension or "bld" in extension:
        proper_method = "excel"
    else:
        proper_method = extension.replace(".", "")
        with open(metadata_path, "r") as infile:
            first_line = infile.read()
            # check for separators in line
            separators = ["\t", ","]
            for sep in separators:
                if sep in first_line:
                    separators_present.append(sep)

    try:
        warnings.filterwarnings("ignore", message="ParserWarning: Falling*")
        use_me_to_read = methods.get(proper_method, None)

        if proper_method != "excel":
            if "\t" in separators_present:
                separator = "\t"
            else:
                separator = ","
            metadata_dataframe = use_me_to_read(metadata_path, sep=separator)
        else:
            metadata_dataframe = use_me_to_read(metadata_path, sheet_name=None)
            # check to see if there are multiple sheets in this input file
            multiple_sheets = pandas.ExcelFile(metadata_path).sheet_names
            first_sheet = multiple_sheets.pop(0)
            if len(multiple_sheets) >= 1:
                for index, sheet_name in enumerate(multiple_sheets):
                    for column in metadata_dataframe[sheet_name].columns:
                        metadata_dataframe[first_sheet][column] = metadata_dataframe[
                            sheet_name
                        ][column]

            metadata_dataframe = metadata_dataframe[first_sheet]

    except (IOError, ValueError) as err:
        try:
            metadata_dataframe = pandas.read_csv(
                metadata_path, sep=separator, engine="python"
            )
        except IOError:
            log.error(
                f"Tried falling back to reading {metadata_path} with pandas.read_csv, still unable to parse"
            )
            raise err(f"Problem opening {metadata_path}")

    return metadata_dataframe


def translate_metadata(metadata_path, metadata_translation_script_path, **kwargs):
    log = logger("pypet2bids")
    # load metadata
    metadata_dataframe = open_meta_data(metadata_path)

    if metadata_dataframe is not None:
        try:
            # this is where the goofiness happens, we allow the user to create their own custom script to manipulate
            # data from their particular spreadsheet wherever that file is located.
            spec = importlib.util.spec_from_file_location(
                "metadata_translation_script_path", metadata_translation_script_path
            )
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            # note the translation must have a method named translate metadata in order to work
            text_file_data = module.translate_metadata(metadata_dataframe, **kwargs)
        except AttributeError as err:
            log.warning(f"Unable to locate metadata_translation_script\n{err}")
    else:
        log.info(f"No metadata found at {metadata_path}")
        text_file_data = None

    return text_file_data


def import_and_write_out_module(module: str, destination: str):
    """
    Writes an imported module file to a destination
    :param module: an imported python module
    :param destination: the destination to write the source/script of the module to file

    :return: the destination path of the copied module file if successful
    """
    imported_module = importlib.import_module(module)
    path_to_module = os.path.abspath(imported_module.__file__)
    shutil.copy(path_to_module, destination)
    if os.path.isfile(destination):
        return destination
    elif os.path.isdir(destination):
        return os.path.join(destination, os.path.basename(path_to_module))


def write_out_module(module: str = "pypet2bids.metadata_spreadsheet_example_reader"):
    parser = argparse.ArgumentParser(
        description="[DEPRECATED!!] Write out a template for a python script used for "
        "bespoke metadata."
    )
    parser.add_argument(
        "template_path",
        type=str,
        help="Path to write out template for a translation script.",
    )
    args = parser.parse_args()

    import_and_write_out_module(module=module, destination=args.template_path)


def expand_path(path_like: str) -> str:
    """
    Expands relative ~ paths to full paths

    :param path_like: path like string
    :type path_like: string

    :return: full path string
    :rtype: string
    """
    if path_like:
        if path_like[0] == "~":
            return str(os.path.expanduser(path_like))
        else:
            return os.path.abspath(path_like)
    else:
        return ""


def collect_bids_part(bids_part: str, path_like: Union[str, pathlib.Path]) -> str:
    """
    Regex is hard, this finds a bids key if it's present in path or pathlink string

    >>> bids_like_path = '/home/users/user/bids_data/sub-NDAR123/ses-firstsession'
    >>> subject_id = collect_bids_part('sub', bids_like_path)
    >>> subject_id
    >>> "sub-NDAR123"

    :param bids_part: the bids part to find in the path e.g. subject id, session id, recording, etc etc
    :type bids_part: string
    :param path_like: a pathlike input, this is strictly for parsing file paths or file like strings
    :type path_like: string or pathlib.Path object, we don't discriminate
    :return: the collected bids part
    :rtype: string
    """
    log = logger("pypet2bids")

    # get os of system
    if os.name == "posix":
        not_windows = True
    else:
        not_windows = False

    # break up path into parts
    parts = pathlib.PurePath(path_like).parts

    # this shouldn't happen, but we check if someone passed a windows path to a posix machine
    # should we check for the inverse? No, not until someone complains about this loudly enough.
    for part in parts:
        if "\\" in part and not_windows:
            # explicitly use windows path splitting
            parts = pathlib.PureWindowsPath(path_like).parts
            log.warning(
                f"Detected \\ in BIDS like path {path_like}, but os is {os.name}, doing best to parse."
            )
            break

    # create search string
    search_string = bids_part + "-(.*)"
    # collect bids_part
    for part in parts:
        found_part = re.search(search_string, part)
        if found_part:
            collected_part = found_part[0]
            break
        else:
            collected_part = ""

    if "_" in collected_part:
        parts = collected_part.split("_")
        for part in parts:
            found_part = re.search(search_string, part)
            if found_part:
                collected_part = found_part[0]
                break
            else:
                collected_part = ""

    return collected_part


def get_coordinates_containing(
    containing: typing.Union[str, int, float],
    dataframe: pandas.DataFrame,
    exact: bool = False,
    single=False,
) -> typing.Union[list, tuple]:
    """
    Collects the coordinates (row and column) containing an input value, that value could be a string, integer, or
    float. When searching for integers or floats it is most likely best to use set the exact argument to True; the same
    goes for finding the exact match of a string. This method is primarily meant to find the row corresponding to a
    subject ID, e.g.

    if your dataframe is given as

    subject_id    some_values    more_values
    sub-NDAR1     1              2
    sub-NDAR2     3              3

    And you provided the arguments: `containing='sub-NDAR', exact=False` you would return a set of the following
    coordinates

    [(0, 'subject_id'), (1, 'subject_id')]

    If you are confident in your input data you would most likely call this method this way:

    >>> coordinates = get_coordinates_containing(
    >>>    'sub-NDAR1',
    >>>    pandas.DataFrame({'subject_id': ['sub-NDAR1', 'sub-NDAR2'], 'some_values': [1, 2]}),
    >>>    single=True)
    >>> coordinates
    >>> (0, 1)

    :param containing: value you wish to search a dataframe for
    :type containing: string, integer, or float
    :param dataframe: A pandas dataframe read in from a spreadsheet
    :type dataframe: pandas.datafarame
    :param exact: Boolean proscribing an exact match to containing; default is to locate a string or value that holds
        the input containing
    :type exact: bool
    :param single: return only the first coordinate, use only if the string/contains your searching for is unique and
        you have high confidence in your data
    :type single: bool
    :return: A coordinate in the form of (row, column) or a list containing a set of coordinates [(row, column), ...]
    :rtype: tuple or list of tuples

    """

    percent_tolerance = 0.001
    coordinates = []
    for row_index, row in dataframe.iterrows():
        for column_index, value in row.items():
            if value is not pandas.NA:
                if exact:
                    if value == containing:
                        coordinates.append((row_index, column_index))
                elif (
                    not exact
                    and not isinstance(value, str)
                    and not isinstance(containing, str)
                ):
                    if numpy.isclose(value, containing, rtol=percent_tolerance):
                        coordinates.append((row_index, column_index))
                elif not exact and type(value) is str:
                    if str(containing) in value:
                        coordinates.append((row_index, column_index))
    if single and coordinates:
        return coordinates[0]
    else:
        return coordinates


def transform_row_to_dict(
    row: typing.Union[int, str, pandas.Series], dataframe: pandas.DataFrame = None
) -> dict:
    """
    Parses a row of a dataframe (or a series from a dataframe) into a dictionary, special care is taken to transform
    array like data contained in a single cell to a list of numbers or strings.
    :param row: either the row index as an integer or a pandas.Series object
    :type row: int, pandas.Series
    :param dataframe: Provided if only an int is given to this method, will be queried at the row by the given integer
    :type dataframe: pandas.DataFrame
    :return: a dictionary corresponding a row from a (most likely multi-subject input sheet) spreadsheet
    :rtype: dict
    """
    if type(row) is pandas.Series:
        row = row
    elif type(row) is str or type(row) is int and dataframe is not None:
        if type(row) is int:
            row = dataframe.iloc[row]
        else:
            row = dataframe.loc[row]

    transformed = {}
    for key, value in row.items():
        try:
            evaluated = ast.literal_eval(str(value))
            if type(evaluated) is tuple:
                evaluated = list(evaluated)
        except (SyntaxError, ValueError):
            evaluated = value
        transformed[key] = evaluated

    return transformed


# noinspection PyPep8Naming
def get_recon_method(ReconstructionMethodString: str) -> dict:
    """
    Given the reconstruction method from a dicom header this function does its best to determine the name of the
    reconstruction, the number of iterations used in the reconstruction, and the number of subsets in the
    reconstruction.

    :param ReconstructionMethodString:
    :return: dictionary containing PET BIDS fields ReconMethodName, ReconMethodParameterUnits,
        ReconMethodParameterLabels, and ReconMethodParameterValues

    """
    contents = ReconstructionMethodString.replace(" ", "")
    subsets = None
    iterations = None

    # determine order of recon iterations and subsets, this is not  a surefire way to determine this...
    iter_sub_combos = {
        "iter_first": [
            r"\d\di\ds",
            r"\d\di\d\ds",
            r"\di\ds",
            r"\di\d\ds",
            r"i\d\ds\d",
            r"i\d\ds\d\d",
            r"i\ds\d",
            r"i\ds\d\d",
        ],
        "sub_first": [
            r"\d\ds\di",
            r"\d\ds\d\di",
            r"\ds\di",
            r"\ds\d\di",
            r"s\d\di\d",
            r"s\d\di\d\d",
            r"s\di\d",
            r"s\di\d\d",
        ],
    }

    iter_sub_combos["iter_first"] = [
        re.compile(regex) for regex in iter_sub_combos["iter_first"]
    ]
    iter_sub_combos["sub_first"] = [
        re.compile(regex) for regex in iter_sub_combos["sub_first"]
    ]
    order = None
    possible_iter_sub_strings = []
    iteration_subset_string = None
    # run through possible combinations of iteration substitution strings in iter_sub_combos
    for key, value in iter_sub_combos.items():
        for expression in value:
            iteration_subset_string = expression.search(contents)
            if iteration_subset_string:
                order = key
                iteration_subset_string = iteration_subset_string[0]
                possible_iter_sub_strings.append(iteration_subset_string)
    # if matches get ready to pick one
    if possible_iter_sub_strings:
        # sorting matches by string length as our method can return more than one match e.g. 3i21s will return
        # 3i21s and 3i1s or something similar
        possible_iter_sub_strings.sort(key=len)
        # picking the longest match as it's most likely the correct one
        iteration_subset_string = possible_iter_sub_strings[-1]

    # after we've captured the subsets and iterations we next need to separate them out from each other
    if iteration_subset_string and order:
        #  remove all chars replace with spaces
        just_digits = re.sub(r"[a-zA-Z]", " ", iteration_subset_string)
        just_digits = just_digits.strip()
        # split up subsets and iterations
        just_digits = just_digits.split(" ")
        # assign digits to either subsets or iterations based on order information obtained earlier
        if order == "iter_first" and len(just_digits) == 2:
            iterations = int(just_digits[0])
            subsets = int(just_digits[1])
        elif len(just_digits) == 2:
            iterations = int(just_digits[1])
            subsets = int(just_digits[0])
        else:
            # if we don't have both we decide that we don't have either, flawed but works for the samples in
            # test_dcm2niix4pet.py may. Will be updated when non-conforming data is obtained
            pass  # do nothing, this case shouldn't fire.....

    if iteration_subset_string:
        ReconMethodName = re.sub(iteration_subset_string, "", contents)
    else:
        ReconMethodName = contents

    # cleaning up weird chars at end or start of name
    ReconMethodName = re.sub(r"[^a-zA-Z0-9]$", "", ReconMethodName)
    ReconMethodName = re.sub(r"^[^a-zA-Z0-9]", "", ReconMethodName)

    expanded_name = ""
    # get the dimension as it's often somewhere in the name
    dimension = ""

    search_criteria = r"[1-4][D-d]"
    if re.search(search_criteria, ReconMethodName):
        dimension = re.search(search_criteria, ReconMethodName)[0]

    # doing some more manipulation of the recon method name to expand it from not so helpful acronyms
    possible_names = metadata.PET_reconstruction_methods.get("reconstruction_names", [])

    # we want to sort the possible names by longest first that we don't break up an acronym prematurely
    sorted_df = pandas.DataFrame(possible_names).sort_values(
        by="value", key=lambda x: x.str.len(), ascending=False
    )

    possible_names = []
    for row in sorted_df.iterrows():
        possible_names.append({"value": row[1]["value"], "name": row[1]["name"]})

    for name in possible_names:
        if name["value"] in ReconMethodName:
            expanded_name += name["name"] + " "
            ReconMethodName = re.sub(name["value"], "", ReconMethodName)

    if expanded_name != "":
        ReconMethodName = dimension + " " * len(dimension) + expanded_name.rstrip()
        ReconMethodName = " ".join(ReconMethodName.split())

    if ReconMethodName in ["Filtered Back Projection", "3D Reprojection"]:
        ReconMethodParameterLabels = []
        ReconMethodParameterUnits = []
        ReconMethodParameterValues = []
    else:  # assume it is OSEM or a variant
        ReconMethodParameterLabels = ["subsets", "iterations"]
        ReconMethodParameterUnits = ["none", "none"]
        ReconMethodParameterValues = [subsets, iterations]

    reconstruction_dict = {
        "ReconMethodName": ReconMethodName,
        "ReconMethodParameterUnits": ReconMethodParameterUnits,
        "ReconMethodParameterLabels": ReconMethodParameterLabels,
        "ReconMethodParameterValues": ReconMethodParameterValues,
    }

    if None in reconstruction_dict["ReconMethodParameterValues"]:
        reconstruction_dict.pop("ReconMethodParameterValues")
        reconstruction_dict.pop("ReconMethodParameterUnits")
        for i in range(len(reconstruction_dict["ReconMethodParameterLabels"])):
            reconstruction_dict["ReconMethodParameterLabels"][i] = "none"

    return reconstruction_dict


def modify_config_file(var: str, value: Union[pathlib.Path, str], config_path=None):
    """
    Given a variable name and a value updates the config file with those inputs.
    Namely used (on Windows, but not limited to) to point to a dcm2niix executable (dcm2niix.exe)
    file as we don't assume dcm2niix is in the path.
    :param var: variable name
    :type var: str
    :param value: variable value, most often a path to another file
    :type value: Union[pathlib.Path, str]
    :param config_path: path to the config file, if not provided this function will look for a file at the user's home
    :type config_path: Union[pathlib.Path, str]
    :return: None
    :rtype: None
    """
    # load dcm2niix file
    if not config_path:
        config_file = pathlib.Path.home() / ".pet2bidsconfig"
    else:
        config_file = pathlib.Path(config_path)

    if config_file.exists():
        # open the file and read in all the lines
        temp_file = config_file.with_suffix(".temp")
        with open(config_file, "r") as infile, open(temp_file, "w") as outfile:
            updated_file = False
            for line in infile:
                if var + "=" in line:
                    outfile.write(f"{var}={value}\n")
                    updated_file = True
                else:
                    outfile.write(line)
            if not updated_file:
                outfile.write(f"{var}={value}\n")
        if system().lower() == "windows":
            config_file.unlink(missing_ok=True)
        temp_file.replace(config_file)
    else:
        # create the file
        try:
            with open(config_file, "w") as outfile:
                outfile.write(f"{var}={value}\n")
        except FileNotFoundError as err:
            logging.error(f"Unable to write to {config_file}\n{err}")


def check_units(entity_key: str, entity_value: str, accepted_units: Union[list, str]):
    """
    Given an entity's name, value, and an accepted range of values (accepted units), check whether those units are
    valid/in accepted units. Raises warning and returns False if units in entity value don't intersect with any entry
    in accepted_units

    :param entity_key: key/name of the entity
    :type entity_key: str
    :param entity_value: units value; some sort of SI unit(s) see (GBq, g, etc)
    :type entity_value: str
    :param accepted_units: a range of accepted units for the BIDS entity
    :type accepted_units: str or list
    :return: Whether the units in entity value ar allowed or not
    :rtype: bool
    """
    allowed = False
    for accepted in accepted_units:
        if entity_value.lower() == accepted.lower:
            allowed = True
            break

    if allowed:
        pass
    else:
        if type(accepted_units) is str or (
            type(accepted_units) is list and len(accepted_units) == 1
        ):
            warning = f"{entity_key} must have units as {accepted_units}, ignoring given units {entity_value}"
        elif type(accepted_units) is list and len(accepted_units) > 1:
            warning = f"{entity_key} must have units as on of  {accepted_units}, ignoring given units {entity_value}"

        logging.warning(warning)

    return allowed


def ad_hoc_checks(
    metadata: dict, modify_input=False, items_that_should_be_checked=None
):
    """
    Check to run on PET BIDS metadata to evaluate whether input is acceptable or not, this function will most likely be
    refactored to use the schema instead of relying on hardcoded checks as listed in items_that_should_be_checked
    :param metadata:
    :type metadata:
    :param modify_input:
    :type modify_input:
    :param items_that_should_be_checked: items to check, hardcoded at the moment, but can accept a dict as input
    :type items_that_should_be_checked: dict
    :return:
    :rtype:
    """
    # dictionary of entities and their acceptable units to check
    if items_that_should_be_checked is None:
        items_that_should_be_checked = {}
    hardcoded_items = {
        "InjectedRadioactivityUnits": ["MBq", "mCi"],
        "SpecificRadioactivityUnits": ["Bq/g", "MBq/ug"],
        "InjectedMassUnits": "ug",
        "MolarActivityUnits": "GBq/umolug",
        "MolecularWeightUnits": "g/mol",
    }

    # if none are
    items_that_should_be_checked.update(**hardcoded_items)

    # iterate through ad hoc items_that_should_be_checked that exist in our metadata
    for entity, units in items_that_should_be_checked.items():
        check_input_entity = metadata.get(entity, None)
        if check_input_entity:
            # this will raise a warning if the units aren't acceptable
            units_are_good = check_units(
                entity_key=entity, entity_value=check_input_entity, accepted_units=units
            )

            # this will remove an entity from metadata form dictionary if it's not good
            if modify_input and not units_are_good:
                metadata.pop(entity)

    return metadata


def sanitize_bad_path(bad_path: Union[str, pathlib.Path]) -> Union[str, pathlib.Path]:
    if " " in str(bad_path):
        return f'"{bad_path}"'.rstrip().strip()
    else:
        return bad_path


def drop_row(dataframe: pandas.DataFrame, index: int):
    row = dataframe.loc[index]
    dataframe.drop(index, inplace=True)
    return row


def replace_nones(dictionary):
    json_string = json.dumps(dictionary)
    # sub nulls
    json_fixed = re.sub("null", '"none"', json_string)
    return json.loads(json_fixed)


def check_pet2bids_config(
    variable: str = "DCM2NIIX_PATH", config_path=Path.home() / ".pet2bidsconfig"
):
    """
    Checks the config file at users home /.pet2bidsconfig for the variable passed in,
    defaults to checking for DCM2NIIX_PATH. However, we can use it for anything we like,
    including setting paths to other useful files or "pet2bids" specific variables we'd like
    to access globally, or set as unchanging defaults.

    :param variable: a string variable name to check for in the config file
    :type variable: string
    :param config_path: path to the config file, defaults to $HOME/.pet2bidsconfig
    :type config_path: string or pathlib.Path
    :return: the value of the variable if it exists in the config file
    :rtype: str
    """
    log = logger("pypet2bids")
    # check to see if path to dcm2niix is in .env file
    dcm2niix_path = None
    variable_value = None
    pypet2bids_config = Path(config_path)
    if pypet2bids_config.exists():
        dotenv.load_dotenv(pypet2bids_config)
        variable_value = os.getenv(variable)
    # else we check our environment variables for the variable
    elif os.getenv(variable) and not pypet2bids_config.exists():
        variable_value = os.getenv(variable)
        log.warning(
            f"Found {variable} in environment variables as {variable_value}, but no .pet2bidsconfig file found at {pypet2bids_config}"
        )
    if variable == "DCM2NIIX_PATH":
        # check to see if dcm2niix is on the path at all
        check_on_path = subprocess.run(
            "dcm2niix -h",
            shell=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        if variable_value:
            # check dcm2niix path exists
            dcm2niix_path = Path(variable_value)
            check = subprocess.run(
                f"{dcm2niix_path} -h",
                shell=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            if check.returncode == 0:
                return dcm2niix_path
        elif (
            not variable_value
            and pypet2bids_config.exists()
            and check_on_path.returncode == 0
        ):
            # do nothing
            # log.info(
            #    f"DCM2NIIX_PATH not found in {pypet2bids_config}, but dcm2niix is on $PATH."
            # )
            return None
        else:
            log.error(
                f"Unable to locate dcm2niix executable at {dcm2niix_path.__str__()}"
                f" Set DCM2NIIX_PATH variable in {pypet2bids_config} or export DCM2NIIX_PATH=/path/to/dcm2niix into your environment variables."
            )
            return None
    if variable != "DCM2NIIX_PATH":
        return variable_value

    if not variable_value and not pypet2bids_config.exists():
        log.warning(
            f"Config file not found at {pypet2bids_config}, .pet2bidsconfig file must exist and "
            f"have variable: {variable} and {variable} must be set."
        )


class CustomFormatter(logging.Formatter):
    """
    Custom debugger courtesy of Sergey Pleshakov on stack overflow
    see https://stackoverflow.com/a/56944256
    """

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = (
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    )

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def hash_fields(**fields):
    hash_return_string = ""
    hash_string = ""
    if fields.get("ProtocolName", None):
        hash_return_string += f"{fields.get('ProtocolName')}_"
    keys_we_want = ["ses", "rec", "trc"]
    for key, value in fields.items():
        # sanitize values
        regex = r"[^a-zA-Z0-9]"
        value = re.sub(regex, "", str(value))
        hash_string += f"{value}_"
        if key in keys_we_want:
            hash_return_string += f"{value}_"

    hash_hex = hashlib.md5(hash_string.encode("utf-8")).hexdigest()

    return f"{hash_return_string}{hash_hex}"


def first_middle_last_frames_to_text(
    four_d_array_like_object, output_folder, step_name="_step_name_"
):
    frames = [
        0,
        four_d_array_like_object.shape[-1] // 2,
        four_d_array_like_object.shape[-1] - 1,
    ]
    frames_to_record = []
    for f in frames:
        frames_to_record.append(four_d_array_like_object[:, :, :, f])

    # now collect a single 2d slice from the "middle" of the 3d frames in frames_to_record
    for index, frame in enumerate(frames_to_record):
        numpy.savetxt(
            output_folder / f"{step_name}_frame_{frames[index]}.tsv",
            frames_to_record[index][:, :, frames_to_record[index].shape[2] // 2],
            delimiter="\t",
            fmt="%s",
        )


def reorder_isotope(isotope: str) -> str:
    """
    Reorders the isotope string to be in the format of "isotope""element name"
    :param isotope: isotope string
    :type isotope: str
    :return: reordered isotope string
    :rtype: str
    """
    # remove all non-alphanumeric characters from isotope
    isotope = re.findall(r"[a-zA-Z0-9]+", isotope)
    # combine all elements in isotope into one string
    isotope = "".join(isotope)
    # collect the isotope number from the isotope string
    isotope_num = re.findall(r"\d+", isotope)
    # collect the element name from the isotope string
    element_name = re.findall(r"[a-zA-Z]+", isotope)

    # capitalize the first letter of the element name if the element name's length is <= 2
    if 1 < len(element_name[0]) <= 2:
        e = element_name[0][0].capitalize()
        if len(element_name[0]) == 2:
            e += element_name[0][1].lower()
    elif len(element_name[0]) == 1:
        e = element_name[0].capitalize()

    # special case for 52mMn
    if "".join(element_name).lower() == "mmn" or "".join(element_name).lower() == "mnm":
        e = "mMn"

    # reorder to put the number before the element name
    isotope = f"{isotope_num[0]}{e}"

    return isotope
